from mpfmc.tests.MpfMcTestCase import MpfMcTestCase


class TestText(MpfMcTestCase):
    def get_machine_path(self):
        return 'tests/machine_files/text_input'

    def get_config_file(self):
        return 'text_input.yaml'

    def test_text_input(self):
        self.mc.events.post('slide1')
        self.advance_time(1)

        text_input_widget = (
            self.mc.targets['default'].current_slide.widgets[1].widget)

        text_display_widget = (
            self.mc.targets['default'].current_slide.widgets[2].widget)

        self.assertEqual(text_display_widget.text, '')
        self.assertEqual(text_input_widget.text, 'C')
        self.advance_time()

        self.mc.events.post('sw_left_flipper')
        self.advance_time()
        self.assertEqual(text_input_widget.text, 'B')
        self.mc.events.post('sw_left_flipper')
        self.advance_time()
        self.assertEqual(text_input_widget.text, 'A')
        self.mc.events.post('sw_left_flipper')
        self.advance_time()
        self.assertEqual(text_input_widget.text, 'END')
        self.mc.events.post('sw_left_flipper')
        self.advance_time()
        self.assertEqual(text_input_widget.text, 'BACK')

        # should not crash if we go back with no chars
        self.mc.events.post('sw_start')
        self.advance_time()

        self.mc.events.post('sw_left_flipper')
        self.advance_time()
        self.assertEqual(text_input_widget.text, 'SPACE')
        self.mc.events.post('sw_left_flipper')
        self.advance_time()
        self.assertEqual(text_input_widget.text, '-')
        self.mc.events.post('sw_left_flipper')
        self.advance_time()
        self.assertEqual(text_input_widget.text, '_')

        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.assertEqual(text_input_widget.text, '-')

        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.assertEqual(text_input_widget.text, 'SPACE')

        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.assertEqual(text_input_widget.text, 'BACK')

        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.assertEqual(text_input_widget.text, 'END')

        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.assertEqual(text_input_widget.text, 'A')

        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.assertEqual(text_input_widget.text, 'B')

        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.assertEqual(text_input_widget.text, 'C')

        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.assertEqual(text_input_widget.text, 'D')

        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.assertEqual(text_input_widget.text, 'E')

        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.assertEqual(text_input_widget.text, 'F')

        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.mc.events.post('sw_right_flipper')
        self.advance_time()

        self.assertEqual(text_input_widget.text, 'K')

        self.mc.events.post('sw_left_flipper')
        self.advance_time()

        self.assertEqual(text_input_widget.text, 'J')

        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.mc.events.post('sw_right_flipper')
        self.advance_time()

        self.assertEqual(text_input_widget.text, 'M')

        self.mc.events.post('sw_start')
        self.advance_time()

        self.assertEqual(text_input_widget.text, 'M')
        self.assertEqual(text_display_widget.text, 'M')

        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.mc.events.post('sw_start')
        self.advance_time()

        self.assertEqual(text_input_widget.text, 'Q')
        self.assertEqual(text_display_widget.text, 'MQ')

        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.mc.events.post('sw_right_flipper')
        self.advance_time()

        self.assertEqual(text_input_widget.text, 'BACK')

        self.mc.events.post('sw_start')
        self.advance_time()

        self.assertEqual(text_input_widget.text, 'Q')
        self.assertEqual(text_display_widget.text, 'M')

        self.mc.events.post('sw_left_flipper')
        self.advance_time()
        self.mc.events.post('sw_start')
        self.advance_time()

        self.assertEqual(text_input_widget.text, 'P')
        self.assertEqual(text_display_widget.text, 'MP')

        self.mc.events.post('sw_left_flipper')
        self.advance_time()
        self.mc.events.post('sw_left_flipper')
        self.advance_time()
        self.mc.events.post('sw_left_flipper')
        self.advance_time()
        self.mc.events.post('sw_left_flipper')
        self.advance_time()
        self.mc.events.post('sw_left_flipper')
        self.advance_time()
        self.mc.events.post('sw_left_flipper')
        self.advance_time()
        self.mc.events.post('sw_left_flipper')
        self.advance_time()
        self.mc.events.post('sw_left_flipper')
        self.advance_time()
        self.mc.events.post('sw_left_flipper')
        self.advance_time()
        self.mc.events.post('sw_left_flipper')
        self.advance_time()

        self.mc.events.post('sw_start')
        self.advance_time()

        self.assertEqual(text_input_widget.text, 'END')
        self.mc.events.post('sw_start')
        self.advance_time()
        self.assertEqual(text_display_widget.text, 'MPF')

    def test_text_input_blocking(self):
        self.mc.events.post('slide1')
        self.advance_time(1)

        text_input_widget = (
            self.mc.targets['default'].current_slide.widgets[1].widget)

        text_display_widget = (
            self.mc.targets['default'].current_slide.widgets[2].widget)

        self.assertEqual(text_display_widget.text, '')
        self.assertEqual(text_input_widget.text, 'C')
        self.advance_time()

        self.mc.events.post('sw_left_flipper')
        self.advance_time()
        self.assertEqual(text_input_widget.text, 'B')
        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.assertEqual(text_input_widget.text, 'C')
        self.mc.events.post('test_block')
        self.advance_time()
        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.assertEqual(text_input_widget.text, 'C')
        self.mc.events.post('sw_left_flipper')
        self.advance_time()
        self.assertEqual(text_input_widget.text, 'C')
        self.mc.events.post('test_release')
        self.advance_time()
        self.mc.events.post('sw_right_flipper')
        self.advance_time()
        self.assertEqual(text_input_widget.text, 'D')
        self.mc.events.post('sw_left_flipper')
        self.advance_time()
        self.assertEqual(text_input_widget.text, 'C')



from mpfmc.tests.MpfMcTestCase import MpfMcTestCase


class TestColor(MpfMcTestCase):
    def get_machine_path(self):
        return 'tests/machine_files/color'

    def get_config_file(self):
        return 'test_color.yaml'

    def get_widget(self, index=0):
        return self.mc.targets['default'].current_slide.children[index].widget

    def test_colors(self):
        self.mc.events.post('slide1')
        self.advance_time()

        self.assertEqual(self.get_widget().color, [1.0, 0, 0, 1.0])
        self.assertEqual(self.get_widget(1).color[0], 0)
        self.assertEqual(self.get_widget(1).color[1], 0)
        self.assertEqual(self.get_widget(1).color[2], 1.0)
        self.assertAlmostEqual(self.get_widget(1).color[3], 0.5, delta=0.1)
        self.assertEqual(self.get_widget(2).color, [0, 1.0, 0, 1.0])



from importlib import import_module

from mpf.core.device_manager import DeviceCollection
from mpf.exceptions.config_file_error import ConfigFileError

MYPY = False
if MYPY:    # pragma: no cover
    from mpfmc.core.mc import MpfMc     # pylint: disable-msg=cyclic-import,unused-import


class ConfigCollection(DeviceCollection):
    """ A lightweight collection of validated configs from the machine or
    mode config. Used to hold configs for things like slides, widgets,
    animations, widget_styles, etc.

    Getter attributes and methods automatically return a deepcopied version of
    the config dict.

    """

    __slots__ = []

    def __init__(self, machine, collection, config_section):
        super().__init__(machine, collection, config_section)
        self.machine.events.add_handler('init_phase_1',
                                        self.create_entries_from_root_config)

        self.machine.events.add_handler('init_phase_2',
                                        self.validate_entries_from_root_config)

        self.machine.mode_controller.register_load_method(
            self.create_entries, self.config_section, priority=100)

        self._initialize()

    @property
    def mc(self) -> "MpfMc":
        return self.machine

    def _initialize(self):
        pass

    def create_entries_from_root_config(self, **kwargs):
        del kwargs
        if self.config_section in self.machine.machine_config:
            self.create_entries(self.machine.machine_config[self.config_section])

    def create_entries(self, config, **kwargs):
        # config is localized to this collection's section
        del kwargs

        for name, settings in config.items():
            # if not settings:
            #     raise AssertionError("{} entry '{}' has an empty config."
            #                          .format(self.config_section, name))
            try:
                self[name] = self.process_config(settings)
            except ConfigFileError as e:
                e.extend("Error creating {} config entry for '{}'".format(self.config_section, name))
                raise e

    def validate_entries_from_root_config(self, **kwargs):
        del kwargs
        if self.config_section in self.machine.machine_config:
            self.validate_entries(self.machine.machine_config[self.config_section])

    def validate_entries(self, config, **kwargs):
        del kwargs
        for name in config.keys():
            self.validate_config(self[name])

    def process_config(self, config):
        raise NotImplementedError

    def validate_config(self, config):
        pass


def create_config_collections(mc, collections):

    for module in collections.values():
        imported_module = import_module(module)
        setattr(mc, imported_module.collection_cls.collection,
                imported_module.collection_cls(mc, imported_module.collection_cls.collection,
                                               imported_module.collection_cls.collection))



from kivy.animation import Animation


class RelativeAnimation(Animation):

    """Class that extends the Kivy Animation base class to add relative animation property target values.

    Those are calculated when the animation starts.
    """

    def _initialize(self, widget):
        """Initializes the animation and calculates the property target value.

        Based on the current value plus the desired delta.

        Notes: Do not call the base class _initialize method as this override
        completely replaces the base class method."""
        d = self._widgets[widget.uid] = {
            'widget': widget,
            'properties': {},
            'time': None}

        # get current values and calculate target values
        p = d['properties']
        for key, value in self._animated_properties.items():
            original_value = getattr(widget, key)
            if isinstance(original_value, (tuple, list)):
                original_value = original_value[:]
                target_value = [x + y for x, y in zip(original_value, value)]
            elif isinstance(original_value, dict):
                original_value = original_value.copy()
                target_value = value
            else:
                target_value = original_value + value
            p[key] = (original_value, target_value)

        # install clock
        self._clock_install()



from mpf.tests.MpfTestCase import MpfTestCase
from mpfmc.widgets.display import DisplayWidget
from mpfmc.uix.widget import WidgetContainer

MYPY = False
if MYPY:   # pragma: no cover
    from mpfmc.uix.slide import Slide


class MpfSlideTestCase(MpfTestCase):

    def assertSlideOnTop(self, slide_name: str, target: str="default"):
        if not self.mc.targets[target].current_slide:
            self.fail("There is no slide on target {}".format(target))
        self.assertEqual(slide_name, self.mc.targets[target].current_slide.name)

    def assertTextOnTopSlide(self, text: str, target: str="default"):
        if not self.mc.targets[target].current_slide:
            self.fail("There is no slide on target {}".format(target))
        self.assertTextInSlide(text, self.mc.targets[target].current_slide.name)

    def assertTextNotOnTopSlide(self, text: str, target: str="default"):
        if not self.mc.targets[target].current_slide:
            return
        self.assertTextNotInSlide(text, self.mc.targets[target].current_slide.name)

    def assertSlideActive(self, slide_name: str):
        self.assertIn(slide_name, self.mc.active_slides, "Slide {} is not active.".format(slide_name))

    def assertSlideNotActive(self, slide_name: str):
        self.assertNotIn(slide_name, self.mc.active_slides, "Slide {} is active but should not.".format(slide_name))

    def _get_texts_from_slide(self, slide: "Slide"):
        texts = []
        for child in slide.children:
            if isinstance(child, WidgetContainer) and child.widget and \
                    isinstance(child.widget, DisplayWidget) and child.widget.current_slide:
                texts.extend(self._get_texts_from_slide(child.widget.current_slide))
            if isinstance(child, WidgetContainer) and child.widget and hasattr(child.widget, "text"):
                texts.append(child.widget.text)

        return texts

    def assertTextInSlide(self, text: str, slide_name: str):
        self.assertSlideActive(slide_name)
        self.assertIn(text, self._get_texts_from_slide(self.mc.active_slides[slide_name]),
                "Text {} not found in slide {}. Text found: {}".format(text, slide_name, self._get_texts_from_slide(self.mc.active_slides[slide_name])))

    def assertTextNotInSlide(self, text: str, slide_name: str):
        self.assertSlideActive(slide_name)
        self.assertNotIn(text, self._get_texts_from_slide(self.mc.active_slides[slide_name]),
                         "Text {} found in slide {} but should not be there.".format(text, slide_name))



from typing import Union, List
from functools import partial
from importlib import import_module

from mpf.core.case_insensitive_dict import CaseInsensitiveDict
from mpf.core.utility_functions import Util
from mpf.core.placeholder_manager import TextTemplate
from mpfmc.config_collections.animation import AnimationCollection
from mpfmc.core.config_collection import ConfigCollection
from mpfmc.uix.widget import magic_events

MYPY = False
if MYPY:   # pragma: no cover
    from mpfmc.uix.widget import Widget


class WidgetCollection(ConfigCollection):

    config_section = 'widgets'
    collection = 'widgets'
    class_label = 'WidgetConfig'

    type_map = CaseInsensitiveDict()

    def _initialize(self) -> None:
        for cls_name, module in self.mc.machine_config['mpf-mc']['widgets'].items():
            for widget_cls in import_module(module).widget_classes:
                self.type_map[cls_name] = widget_cls

    def process_config(self, config: Union[dict, list]) -> List["Widget"]:
        # config is localized to a specific widget section
        if config is None:
            config = []
        if isinstance(config, dict):
            config = [config]

        widget_list = list()

        for widget in config:
            widget_list.append(self.process_widget(widget))

        return widget_list

    def _load_widget_by_name(self, config) -> dict:
        # Replace placeholders
        if "(" in config['widget'] and ")" in config['widget']:
            config = dict(config)
            config["widget"] = TextTemplate(self.machine, config['widget'].replace("(", "{").replace(")", "}"))

        return config

    def validate_config(self, config):
        """Check that this widget is valid."""
        for widget_config in config:
            if isinstance(widget_config.get("widget"), str) and widget_config['widget'] not in self.mc.widgets:
                raise ValueError('"{}" is not a valid widget name.'.format(widget_config['widget']))

    def process_widget(self, config: dict) -> dict:     # noqa
        if config.get("widget"):
            return self._load_widget_by_name(config)

        # config is localized widget settings
        try:
            widget_cls = WidgetCollection.type_map[config['type']]
        except (KeyError, TypeError):
            try:
                raise ValueError(
                    '"{}" is not a valid MPF display widget type. Did you '
                    'misspell it, or forget to enable it in the "mpf-mc: '
                    'widgets" section of your machine config?'.format(
                        config['type']))
            except (KeyError, TypeError):
                raise ValueError("Invalid widget config: {}".format(config))

        config['_default_settings'] = list()

        for default_setting_name in widget_cls.merge_settings:
            if default_setting_name in config:
                config['_default_settings'].append(default_setting_name)

        self.mc.config_validator.validate_config('widgets:{}'.format(
            config['type']).lower(), config, base_spec='widgets:common')

        if 'effects' in config and config['type'] == 'display':
            config['effects'] = self.mc.effects_manager.validate_effects(config['effects'])

        if 'animations' in config:
            config['animations'] = (
                self.process_animations(config['animations']))

        else:
            config['animations'] = None

        if 'control_events' in config:
            for event_dict in config['control_events']:
                if event_dict["event"] not in magic_events:
                    self.mc.events.add_handler("client_connected", partial(self._register_trigger,
                                                                           event_dict["event"]))

        if 'reset_animations_events' in config:
            for event_name in config['reset_animations_events']:
                if event_name not in magic_events:
                    self.mc.events.add_handler("client_connected", partial(self._register_trigger, event_name))

        if config.get('z', 0) < 0:
            raise ValueError(
                "\nWidget with negative z value in config: {}.\n\nAs of MPF "
                "v0.30.3, negative z: "
                "values are no longer used to put widgets in 'parent' frames. "
                "Instead add a 'target:' setting to the 'widget_player:' entry"
                " and set that to the name of the display target (display or "
                "slide_frame) you want to add this widget to. Note that "
                "'target: default' is valid and will add the widget to the "
                "default display on top of any slides.\n".format(config))

        return config

    def _register_trigger(self, event_name: str, **kwargs) -> None:
        del kwargs
        self.mc.bcp_processor.register_trigger(event=event_name)

    def process_animations(self, config: dict) -> dict:
        # config is localized to the slide's 'animations' section

        for event_name, event_settings in config.items():

            # make sure the event_name is registered as a trigger event so MPF
            # will send those events as triggers via BCP. But we don't want
            # to register magic events since those aren't real MPF events.
            if event_name not in magic_events:
                self.mc.events.add_handler("client_connected", partial(self._register_trigger, event_name))

            # str means it's a list of named animations
            if isinstance(event_settings, str):
                event_settings = Util.string_to_event_list(event_settings)

            # dict means it's a single set of settings for one animation step
            elif isinstance(event_settings, dict):
                event_settings = [event_settings]

            # ultimately we're producing a list of dicts, so build that list
            # as we iterate
            new_list = list()
            for settings in event_settings:
                new_list.append(AnimationCollection.process_animation(settings, self.mc.config_validator))

            config[event_name] = new_list

        return config


collection_cls = WidgetCollection



import logging

from mpfmc.tests.MpfMcTestCase import MpfMcTestCase
from unittest.mock import MagicMock, call, ANY

try:
    from mpfmc.core.audio import SoundSystem
    from mpfmc.assets.sound import SoundInstance, SoundStealingMethod, SoundPool, ModeEndAction
except ImportError:
    SoundSystem = None
    SoundInstance = None
    SoundStealingMethod = None
    ModeEndAction = None
    SoundPool = None
    logging.warning("mpfmc.core.audio library could not be loaded. Audio "
                    "features will not be available")


class TestAudio(MpfMcTestCase):
    """
    Tests all the audio features in the media controller.  The core audio library is a
    custom extension library written in Cython that interfaces with the SDL2 and
    SDL_Mixer libraries.
    """
    def get_machine_path(self):
        return 'tests/machine_files/audio'

    def get_config_file(self):
        return 'test_audio.yaml'

    def test_typical_sound_system(self):
        """ Tests the sound system and audio interface with typical settings """

        if SoundSystem is None or self.mc.sound_system is None:
            log = logging.getLogger('TestAudio')
            log.warning("Sound system is not enabled - skipping audio tests")
            self.skipTest("Sound system is not enabled")

        self.assertIsNotNone(self.mc.sound_system)
        interface = self.mc.sound_system.audio_interface
        if interface is None:
            log = logging.getLogger('TestAudio')
            log.warning("Sound system audio interface could not be loaded - skipping audio tests")
            self.skipTest("Sound system audio interface could not be loaded")

        self.assertIsNotNone(interface)

        # Check basic audio interface settings
        settings = interface.get_settings()
        self.assertIsNotNone(settings)

        # Check static conversion functions (gain, samples)
        self.assertEqual(interface.string_to_gain('0db'), 1.0)
        self.assertAlmostEqual(interface.string_to_gain('-3 db'), 0.707945784)
        self.assertAlmostEqual(interface.string_to_gain('-6 db'), 0.501187233)
        self.assertAlmostEqual(interface.string_to_gain('-17.5 db'), 0.133352143)
        self.assertEqual(interface.string_to_gain('3db'), 1.0)
        self.assertEqual(interface.string_to_gain('0.25'), 0.25)
        self.assertEqual(interface.string_to_gain('-3'), 0.0)

        self.assertEqual(interface.string_to_samples("234"), 234)
        self.assertEqual(interface.string_to_samples("234.73"), 234)
        self.assertEqual(interface.string_to_samples("-23"), -23)
        self.assertEqual(interface.string_to_samples("2s"), 88200)
        self.assertEqual(interface.string_to_samples("2 ms"), 88)
        self.assertEqual(interface.string_to_samples("23.5 ms"), 1036)
        self.assertEqual(interface.string_to_samples("-2 ms"), -88)

        self.assertEqual(interface.convert_seconds_to_buffer_length(2.25), 396900)
        self.assertEqual(interface.convert_buffer_length_to_seconds(396900), 2.25)

        # Check tracks
        self.assertEqual(interface.get_track_count(), 3)
        track_voice = interface.get_track_by_name("voice")
        self.assertIsNotNone(track_voice)
        self.assertEqual(track_voice.name, "voice")
        self.assertAlmostEqual(track_voice.volume, 0.6, 1)
        self.assertEqual(track_voice.max_simultaneous_sounds, 1)

        track_sfx = interface.get_track_by_name("sfx")
        self.assertIsNotNone(track_sfx)
        self.assertEqual(track_sfx.name, "sfx")
        self.assertAlmostEqual(track_sfx.volume, 0.4, 1)
        self.assertEqual(track_sfx.max_simultaneous_sounds, 8)

        track_music = interface.get_track_by_name("music")
        self.assertIsNotNone(track_music)
        self.assertEqual(track_music.name, "music")
        self.assertAlmostEqual(track_music.volume, 0.5, 1)
        self.assertEqual(track_music.max_simultaneous_sounds, 1)

        self.assertTrue(hasattr(self.mc, 'sounds'))

        # Mock BCP send method
        self.mc.bcp_processor.send = MagicMock()
        self.mc.bcp_processor.enabled = True

        # Allow some time for sound assets to load
        self.advance_real_time(2)

        # Start mode
        self.send(bcp_command='mode_start', name='mode1', priority=500)
        self.assertTrue(self.mc.modes['mode1'].active)
        self.assertEqual(self.mc.modes['mode1'].priority, 500)

        # /sounds/sfx
        self.assertIn('198361_sfx-028', self.mc.sounds)     # .wav
        self.assertIn('210871_synthping', self.mc.sounds)   # .wav
        self.assertIn('264828_text', self.mc.sounds)        # .ogg
        self.assertIn('4832__zajo__drum07', self.mc.sounds)   # .wav
        self.assertIn('84480__zgump__drum-fx-4', self.mc.sounds)   # .wav
        self.assertIn('100184__menegass__rick-drum-bd-hard', self.mc.sounds)   # .wav

        # /sounds/voice
        self.assertIn('104457_moron_test', self.mc.sounds)  # .wav
        self.assertEqual(self.mc.sounds['104457_moron_test'].volume, 0.6)
        self.assertIn('113690_test', self.mc.sounds)        # .wav

        # /sounds/music
        self.assertIn('263774_music', self.mc.sounds)       # .wav

        # Sound groups
        self.assertIn('drum_group', self.mc.sounds)
        self.assertTrue(isinstance(self.mc.sounds['drum_group'], SoundPool))

        # Make sure sound has ducking (since it was specified in the config files)
        self.assertTrue(self.mc.sounds['104457_moron_test'].has_ducking)
        self.assertEqual(2.0, self.mc.sounds['104457_moron_test'].about_to_finish_time)
        self.assertListEqual(['moron_test_about_to_finish'],
                             self.mc.sounds['104457_moron_test'].events_when_about_to_finish)

        # Test sound_player
        self.assertFalse(track_sfx.sound_is_playing(self.mc.sounds['264828_text']))
        self.mc.events.post('play_sound_text')
        self.mc.events.post('play_sound_music')
        self.advance_real_time(1)
        self.assertTrue(track_sfx.sound_is_playing(self.mc.sounds['264828_text']))

        # Test two sounds at the same time on the voice track (only
        # 1 sound at a time max).  Second sound should be queued and
        # play immediately after the first one ends.
        self.assertEqual(track_voice.get_sound_queue_count(), 0)
        self.mc.events.post('play_sound_test')
        self.advance_real_time()

        # Make sure first sound is playing on the voice track
        self.assertEqual(track_voice.get_status()[0]['sound_id'], self.mc.sounds['113690_test'].id)
        self.mc.events.post('play_sound_moron_test')
        self.advance_real_time()

        # Make sure first sound is still playing and the second one has been queued
        self.assertEqual(track_voice.get_status()[0]['sound_id'], self.mc.sounds['113690_test'].id)
        self.assertEqual(track_voice.get_sound_queue_count(), 1)
        self.assertTrue(track_voice.sound_is_in_queue(self.mc.sounds['104457_moron_test']))
        self.advance_real_time(0.1)

        # Now stop sound that is not yet playing but is queued (should be removed from queue)
        self.mc.events.post('stop_sound_moron_test')
        self.advance_real_time(0.25)
        self.assertFalse(track_voice.sound_is_in_queue(self.mc.sounds['104457_moron_test']))

        # Play moron test sound again (should be added to queue)
        self.mc.events.post('play_sound_moron_test')
        self.advance_real_time(0.1)
        self.assertTrue(track_voice.sound_is_in_queue(self.mc.sounds['104457_moron_test']))

        # Make sure text sound is still playing (looping)
        self.assertTrue(track_sfx.sound_is_playing(self.mc.sounds['264828_text']))

        # Ensure sound.events_when_looping is working properly (send event when a sound loops)
        self.mc.bcp_processor.send.assert_any_call('trigger', sound_instance=ANY, name='text_sound_looping')

        # Send an event to stop the text sound looping
        self.mc.events.post('stop_sound_looping_text')
        self.advance_real_time(2)

        # Text sound should no longer be playing
        self.assertFalse(track_sfx.sound_is_playing(self.mc.sounds['264828_text']))

        self.advance_real_time(2.7)
        self.mc.events.post('play_sound_synthping')
        self.advance_real_time(3)
        self.assertEqual(track_voice.get_status()[0]['sound_id'], self.mc.sounds['104457_moron_test'].id)
        self.assertEqual(track_voice.get_status()[0]['volume'], 76)
        self.mc.events.post('play_sound_synthping')
        self.advance_real_time(6)
        self.mc.events.post('stop_sound_music')
        self.mc.events.post('play_sound_synthping_in_mode')
        self.advance_real_time(1)
        self.mc.events.post('play_sound_synthping')
        self.advance_real_time(1)

        # Test playing sound pool (many times)
        for x in range(16):
            self.mc.events.post('play_sound_drum_group')
            self.advance_real_time(0.1)

        self.mc.events.post('play_sound_drum_group_in_mode')
        self.advance_real_time(1)

        # Test stopping the mode
        self.send(bcp_command='mode_stop', name='mode1')
        self.advance_real_time(1)

        # Test sound events
        self.mc.bcp_processor.send.assert_any_call('trigger', sound_instance=ANY, name='moron_test_played')
        self.mc.bcp_processor.send.assert_any_call('trigger', sound_instance=ANY, name='moron_test_about_to_finish')
        self.mc.bcp_processor.send.assert_any_call('trigger', sound_instance=ANY, name='moron_test_stopped')
        self.mc.bcp_processor.send.assert_any_call('trigger', sound_instance=ANY, name='synthping_played')
        self.mc.bcp_processor.send.assert_any_call('trigger', sound_instance=ANY, marker_id=0, name='moron_marker')
        self.mc.bcp_processor.send.assert_any_call('trigger', sound_instance=ANY, marker_id=1, name='moron_next_marker')
        self.mc.bcp_processor.send.assert_any_call('trigger', sound_instance=ANY, marker_id=1, name='last_marker')
        self.mc.bcp_processor.send.assert_any_call('trigger', sound_instance=ANY, marker_id=2,
                                                   name='moron_about_to_finish_marker')

    def test_sound_instance_parameters(self):
        """Test the creation of sound instances and their parameters"""

        if SoundSystem is None or self.mc.sound_system is None:
            log = logging.getLogger('TestAudio')
            log.warning("Sound system is not enabled - skipping audio tests")
            self.skipTest("Sound system is not enabled")

        self.assertIsNotNone(self.mc.sound_system)
        interface = self.mc.sound_system.audio_interface
        if interface is None:
            log = logging.getLogger('TestAudio')
            log.warning("Sound system audio interface could not be loaded - skipping audio tests")
            self.skipTest("Sound system audio interface could not be loaded")

        self.assertIsNotNone(interface)

        # Ensure sound we are interested in exists
        self.assertIn('210871_synthping', self.mc.sounds)
        self.assertEqual(self.mc.sounds['210871_synthping'].priority, 1)
        self.assertEqual(self.mc.sounds['210871_synthping'].simultaneous_limit, 3)
        self.assertEqual(self.mc.sounds['210871_synthping'].stealing_method, SoundStealingMethod.oldest)
        self.assertEqual(self.mc.sounds['210871_synthping'].events_when_played, ['synthping_played'])
        self.assertIsNone(self.mc.sounds['210871_synthping'].events_when_stopped)
        self.assertEqual(self.mc.sounds['210871_synthping'].max_queue_time, 2.0)
        self.assertEqual(self.mc.sounds['210871_synthping'].mode_end_action, ModeEndAction.stop_looping)

        # Create sound instance with no overridden parameters (all values come from sound)
        instance1 = SoundInstance(self.mc.sounds['210871_synthping'])
        self.assertEqual(instance1.priority, 1)
        self.assertEqual(instance1.simultaneous_limit, 3)
        self.assertEqual(instance1.stealing_method, SoundStealingMethod.oldest)
        self.assertEqual(instance1.events_when_played, ['synthping_played'])
        self.assertIsNone(instance1.events_when_stopped)
        self.assertEqual(instance1.max_queue_time, 2.0)
        self.assertEqual(instance1.mode_end_action, ModeEndAction.stop_looping)

        # Create sound instance with several overridden parameters
        instance2 = SoundInstance(self.mc.sounds['210871_synthping'], None,
                                  {'priority': 5,
                                   'simultaneous_limit': 7,
                                   'stealing_method': 'skip',
                                   'events_when_played': ['use_sound_setting'],
                                   'events_when_stopped': ['synthping_stopped'],
                                   'max_queue_time': None,
                                   'mode_end_action': 'stop'})
        self.assertEqual(instance2.priority, 5)
        self.assertEqual(instance2.events_when_played, ['synthping_played'])
        self.assertEqual(instance2.events_when_stopped, ['synthping_stopped'])
        self.assertIsNone(instance2.max_queue_time)
        self.assertEqual(instance2.mode_end_action, ModeEndAction.stop)

        # These parameters may not be overridden (supplied values are ignored)
        self.assertEqual(instance2.simultaneous_limit, 3)
        self.assertEqual(instance2.stealing_method, SoundStealingMethod.oldest)

        # Create sound instance with several overridden parameters
        instance3 = SoundInstance(self.mc.sounds['210871_synthping'], None,
                                  {'priority': None,
                                   'events_when_played': None,
                                   'events_when_stopped': ['use_sound_setting'],
                                   'max_queue_time': 0.0})
        self.assertEqual(instance3.priority, 1)
        self.assertIsNone(instance3.events_when_played)
        self.assertIsNone(instance3.events_when_stopped)
        self.assertEqual(instance3.max_queue_time, 0.0)

    def test_mode_sounds(self):
        """ Test the sound system using sounds specified in a mode """

        if SoundSystem is None or self.mc.sound_system is None:
            log = logging.getLogger('TestAudio')
            log.warning("Sound system is not enabled - skipping audio tests")
            self.skipTest("Sound system is not enabled")

        self.assertIsNotNone(self.mc.sound_system)
        interface = self.mc.sound_system.audio_interface
        if interface is None:
            log = logging.getLogger('TestAudio')
            log.warning("Sound system audio interface could not be loaded - skipping audio tests")
            self.skipTest("Sound system audio interface could not be loaded")

        self.assertIsNotNone(interface)

        self.assertTrue(self.mc, 'sounds')

        # Mock BCP send method
        self.mc.bcp_processor.send = MagicMock()
        self.mc.bcp_processor.enabled = True

        # Get tracks
        track_sfx = interface.get_track_by_name("sfx")
        self.assertIsNotNone(track_sfx)
        self.assertEqual(track_sfx.name, "sfx")
        track_music = interface.get_track_by_name("music")
        self.assertIsNotNone(track_music)
        self.assertEqual(track_music.name, "music")

        self.assertIn('263774_music', self.mc.sounds)  # .wav
        music_sound = self.mc.sounds['263774_music']
        music_sound.load()

        # Allow some time for sound assets to load
        self.advance_real_time(2)

        # Start mode
        self.send(bcp_command='mode_start', name='mode2', priority=500)
        self.assertTrue(self.mc.modes['mode2'].active)
        self.assertEqual(self.mc.modes['mode2'].priority, 500)
        self.assertIn('boing_mode2', self.mc.sounds)  # .wav
        boing_sound = self.mc.sounds['boing_mode2']
        self.advance_real_time()

        # Play a longer music sound (launched by sound_player in mode2).  Will test
        # sound's mode_end_action when the mode ends.
        self.mc.events.post('play_sound_boing_in_mode2')
        self.advance_real_time()
        self.assertTrue(track_sfx.sound_is_playing(boing_sound))
        self.advance_real_time(1)

        self.mc.events.post('play_sound_music_fade_at_mode_end')
        self.advance_real_time(0.25)
        self.assertTrue(track_music.sound_is_playing(music_sound))
        self.advance_real_time(2)

        # End mode (sound should fade out for 1 second then stop)
        self.send(bcp_command='mode_stop', name='mode2')
        self.assertFalse(self.mc.modes['mode2'].active)
        self.advance_real_time()
        self.assertTrue(track_music.sound_is_playing(music_sound))
        self.advance_real_time(1.1)
        self.assertFalse(track_music.sound_is_playing(music_sound))

    def test_sound_fading(self):
        """ Tests the fading of sounds"""

        if SoundSystem is None or self.mc.sound_system is None:
            log = logging.getLogger('TestAudio')
            log.warning("Sound system is not enabled - skipping audio tests")
            self.skipTest("Sound system is not enabled")

        self.assertIsNotNone(self.mc.sound_system)
        interface = self.mc.sound_system.audio_interface
        if interface is None:
            log = logging.getLogger('TestAudio')
            log.warning("Sound system audio interface could not be loaded - skipping audio tests")
            self.skipTest("Sound system audio interface could not be loaded")

        self.assertIsNotNone(interface)

        track_music = interface.get_track_by_name("music")
        self.assertIsNotNone(track_music)
        self.assertEqual(track_music.name, "music")
        self.assertEqual(track_music.max_simultaneous_sounds, 1)

        self.advance_real_time(2)

        self.assertIn('263774_music', self.mc.sounds)       # .wav
        music = self.mc.sounds['263774_music']
        retry_count = 10
        while not music.loaded and retry_count > 0:
            if not music.loading:
                music.load()
            self.advance_real_time(0.5)
            retry_count -= 1

        self.assertTrue(music.loaded)
        instance1 = track_music.play_sound(music, context=None, settings={'fade_in': 2.0, 'volume': 1.0})
        self.advance_real_time()

        status = track_music.get_status()
        self.assertEqual(status[0]['sound_instance_id'], instance1.id)
        self.assertEqual(status[0]['status'], "playing")
        self.assertEqual(status[0]['fading_status'], "fade in")
        self.advance_real_time(2)

        instance1.stop(1)
        self.advance_real_time()
        status = track_music.get_status()
        self.assertEqual(status[0]['status'], "stopping")
        self.assertEqual(status[0]['fading_status'], "fade out")
        self.advance_real_time(1)
        status = track_music.get_status()
        self.assertEqual(status[0]['status'], "idle")

        instance2 = track_music.play_sound(music, context=None, settings={'fade_in': 0, 'volume': 1.0})
        self.advance_real_time(1)

        status = track_music.get_status()
        self.assertEqual(status[0]['sound_instance_id'], instance2.id)
        self.assertEqual(status[0]['status'], "playing")
        self.assertEqual(status[0]['fading_status'], "not fading")

        instance2.stop(0)
        self.advance_real_time(0.5)
        status = track_music.get_status()
        self.assertEqual(status[0]['status'], "idle")

    def test_sound_start_at(self):
        """ Tests starting a sound at a position other than the beginning"""

        if SoundSystem is None or self.mc.sound_system is None:
            log = logging.getLogger('TestAudio')
            log.warning("Sound system is not enabled - skipping audio tests")
            self.skipTest("Sound system is not enabled")

        self.assertIsNotNone(self.mc.sound_system)
        interface = self.mc.sound_system.audio_interface
        if interface is None:
            log = logging.getLogger('TestAudio')
            log.warning("Sound system audio interface could not be loaded - skipping audio tests")
            self.skipTest("Sound system audio interface could not be loaded")

        self.assertIsNotNone(interface)

        track_music = interface.get_track_by_name("music")
        self.assertIsNotNone(track_music)
        self.assertEqual(track_music.name, "music")
        self.assertEqual(track_music.max_simultaneous_sounds, 1)

        self.assertIn('263774_music', self.mc.sounds)  # .wav
        music_sound = self.mc.sounds['263774_music']
        music_sound.load()

        self.advance_real_time(2)

        self.assertTrue(music_sound.loaded)
        settings = {'start_at': 7.382}
        instance = self.mc.sounds['263774_music'].play(settings=settings)
        self.advance_real_time()
        status = track_music.get_status()
        self.assertGreaterEqual(status[0]['sample_pos'], interface.convert_seconds_to_buffer_length(7.382))
        self.advance_real_time(1)
        instance.stop(0.25)
        self.advance_real_time(0.3)

    def test_sound_instance_management(self):
        """ Tests instance management of sounds"""

        if SoundSystem is None or self.mc.sound_system is None:
            log = logging.getLogger('TestAudio')
            log.warning("Sound system is not enabled - skipping audio tests")
            self.skipTest("Sound system is not enabled")

        self.assertIsNotNone(self.mc.sound_system)
        interface = self.mc.sound_system.audio_interface
        if interface is None:
            log = logging.getLogger('TestAudio')
            log.warning("Sound system audio interface could not be loaded - skipping audio tests")
            self.skipTest("Sound system audio interface could not be loaded")

        self.assertIsNotNone(interface)

        # Mock BCP send method
        self.mc.bcp_processor.send = MagicMock()
        self.mc.bcp_processor.enabled = True

        track_sfx = interface.get_track_by_name("sfx")
        self.assertIsNotNone(track_sfx)
        self.assertEqual(track_sfx.name, "sfx")
        self.assertEqual(track_sfx.max_simultaneous_sounds, 8)
        self.advance_real_time()

        # Test skip stealing method
        self.assertIn('264828_text', self.mc.sounds)  # .wav
        text_sound = self.mc.sounds['264828_text']
        if not text_sound.loaded:
            if not text_sound.loading:
                text_sound.load()
            self.advance_real_time(1)

        self.assertEqual(text_sound.simultaneous_limit, 3)
        if SoundStealingMethod is not None:
            self.assertEqual(text_sound.stealing_method, SoundStealingMethod.skip)

        instance1 = text_sound.play(settings={'loops': 0, 'events_when_played': ['instance1_played']})
        instance2 = text_sound.play(settings={'loops': 0, 'events_when_played': ['instance2_played']})
        instance3 = text_sound.play(settings={'loops': 0, 'events_when_played': ['instance3_played']})
        instance4 = text_sound.play(settings={'loops': 0, 'events_when_played': ['instance4_played']})
        instance5 = text_sound.play(settings={'loops': 0, 'events_when_played': ['instance5_played']})

        self.advance_real_time(0.5)

        self.mc.bcp_processor.send.assert_any_call('trigger', sound_instance=ANY, name='instance1_played')
        self.mc.bcp_processor.send.assert_any_call('trigger', sound_instance=ANY, name='instance2_played')
        self.mc.bcp_processor.send.assert_any_call('trigger', sound_instance=ANY, name='instance3_played')
        with self.assertRaises(AssertionError):
            self.mc.bcp_processor.send.assert_any_call('trigger', sound_instance=ANY, name='instance4_played')
            self.mc.bcp_processor.send.assert_any_call('trigger', sound_instance=ANY, name='instance5_played')

        self.assertTrue(instance1.played)
        self.assertTrue(instance2.played)
        self.assertTrue(instance3.played)
        self.assertIsNone(instance4)
        self.assertIsNone(instance5)

        track_sfx.stop_all_sounds()
        self.advance_real_time()

        # Test oldest stealing method
        self.mc.bcp_processor.send.reset_mock()
        self.assertIn('210871_synthping', self.mc.sounds)  # .wav
        synthping = self.mc.sounds['210871_synthping']
        if not synthping.loaded:
            if not synthping.loading:
                synthping.load()
            self.advance_real_time(1)
        self.assertEqual(synthping.simultaneous_limit, 3)
        if SoundStealingMethod is not None:
            self.assertEqual(synthping.stealing_method, SoundStealingMethod.oldest)

        synthping_instance1 = synthping.play(settings={'events_when_played': ['synthping_instance1_played'],
                                                       'events_when_stopped': ['synthping_instance1_stopped']})
        self.advance_real_time()
        self.mc.bcp_processor.send.assert_has_calls([call('trigger', sound_instance=ANY, name='synthping_instance1_played')])

        synthping_instance2 = synthping.play(settings={'events_when_played': ['synthping_instance2_played'],
                                                       'events_when_stopped': ['synthping_instance2_stopped']})
        self.advance_real_time()
        self.mc.bcp_processor.send.assert_has_calls([call('trigger', sound_instance=ANY, name='synthping_instance1_played'),
                                                     call('trigger', sound_instance=ANY, name='synthping_instance2_played')])

        synthping_instance3 = synthping.play(settings={'events_when_played': ['synthping_instance3_played'],
                                                       'events_when_stopped': ['synthping_instance3_stopped']})
        self.advance_real_time()
        self.mc.bcp_processor.send.assert_has_calls([call('trigger', sound_instance=ANY, name='synthping_instance1_played'),
                                                     call('trigger', sound_instance=ANY, name='synthping_instance2_played'),
                                                     call('trigger', sound_instance=ANY, name='synthping_instance3_played')])

        synthping_instance4 = synthping.play(settings={'events_when_played': ['synthping_instance4_played'],
                                                       'events_when_stopped': ['synthping_instance4_stopped']})
        self.advance_real_time()
        self.mc.bcp_processor.send.assert_has_calls([call('trigger', sound_instance=ANY, name='synthping_instance1_played'),
                                                     call('trigger', sound_instance=ANY, name='synthping_instance2_played'),
                                                     call('trigger', sound_instance=ANY, name='synthping_instance3_played'),
                                                     call('trigger', sound_instance=ANY, name='synthping_instance1_stopped'),
                                                     call('trigger', sound_instance=ANY, name='synthping_instance4_played')])

        synthping_instance5 = synthping.play(settings={'events_when_played': ['synthping_instance5_played'],
                                                       'events_when_stopped': ['synthping_instance5_stopped']})
        self.advance_real_time()
        self.mc.bcp_processor.send.assert_has_calls([call('trigger', sound_instance=ANY, name='synthping_instance1_played'),
                                                     call('trigger', sound_instance=ANY, name='synthping_instance2_played'),
                                                     call('trigger', sound_instance=ANY, name='synthping_instance3_played'),
                                                     call('trigger', sound_instance=ANY, name='synthping_instance1_stopped'),
                                                     call('trigger', sound_instance=ANY, name='synthping_instance4_played'),
                                                     call('trigger', sound_instance=ANY, name='synthping_instance2_stopped'),
                                                     call('trigger', sound_instance=ANY, name='synthping_instance5_played')])

        self.assertTrue(synthping_instance1.played)
        self.assertTrue(synthping_instance2.played)
        self.assertTrue(synthping_instance3.played)
        self.assertTrue(synthping_instance4.played)
        self.assertTrue(synthping_instance5.played)

        track_sfx.stop_all_sounds()
        self.advance_real_time()

        # Test newest stealing method
        self.mc.bcp_processor.send.reset_mock()
        self.assertIn('198361_sfx-028', self.mc.sounds)  # .wav
        sfx = self.mc.sounds['198361_sfx-028']
        if not sfx.loaded:
            if not sfx.loading:
                sfx.load()
            self.advance_real_time(1)
        self.assertEqual(sfx.simultaneous_limit, 3)
        if SoundStealingMethod is not None:
            self.assertEqual(sfx.stealing_method, SoundStealingMethod.newest)

        sfx_instance1 = sfx.play(
            settings={'events_when_played': ['sfx_instance1_played'], 'events_when_stopped': ['sfx_instance1_stopped']})
        self.advance_real_time()
        self.mc.bcp_processor.send.assert_has_calls([call('trigger', sound_instance=ANY, name='sfx_instance1_played')])

        sfx_instance2 = sfx.play(
            settings={'events_when_played': ['sfx_instance2_played'], 'events_when_stopped': ['sfx_instance2_stopped']})
        self.advance_real_time()
        self.mc.bcp_processor.send.assert_has_calls([call('trigger', sound_instance=ANY, name='sfx_instance1_played'),
                                                     call('trigger', sound_instance=ANY, name='sfx_instance2_played')])

        sfx_instance3 = sfx.play(
            settings={'events_when_played': ['sfx_instance3_played'], 'events_when_stopped': ['sfx_instance3_stopped']})
        self.advance_real_time()
        self.mc.bcp_processor.send.assert_has_calls([call('trigger', sound_instance=ANY, name='sfx_instance1_played'),
                                                     call('trigger', sound_instance=ANY, name='sfx_instance2_played'),
                                                     call('trigger', sound_instance=ANY, name='sfx_instance3_played')])

        sfx_instance4 = sfx.play(
            settings={'events_when_played': ['sfx_instance4_played'], 'events_when_stopped': ['sfx_instance4_stopped']})
        self.advance_real_time()
        self.mc.bcp_processor.send.assert_has_calls([call('trigger', sound_instance=ANY, name='sfx_instance1_played'),
                                                     call('trigger', sound_instance=ANY, name='sfx_instance2_played'),
                                                     call('trigger', sound_instance=ANY, name='sfx_instance3_played'),
                                                     call('trigger', sound_instance=ANY, name='sfx_instance3_stopped'),
                                                     call('trigger', sound_instance=ANY, name='sfx_instance4_played')])

        sfx_instance5 = sfx.play(
            settings={'events_when_played': ['sfx_instance5_played'], 'events_when_stopped': ['sfx_instance5_stopped']})
        self.advance_real_time()
        self.mc.bcp_processor.send.assert_has_calls([call('trigger', sound_instance=ANY, name='sfx_instance1_played'),
                                                     call('trigger', sound_instance=ANY, name='sfx_instance2_played'),
                                                     call('trigger', sound_instance=ANY, name='sfx_instance3_played'),
                                                     call('trigger', sound_instance=ANY, name='sfx_instance3_stopped'),
                                                     call('trigger', sound_instance=ANY, name='sfx_instance4_played'),
                                                     call('trigger', sound_instance=ANY, name='sfx_instance4_stopped'),
                                                     call('trigger', sound_instance=ANY, name='sfx_instance5_played')])

        self.assertTrue(sfx_instance1.played)
        self.assertTrue(sfx_instance2.played)
        self.assertTrue(sfx_instance3.played)
        self.assertTrue(sfx_instance4.played)
        self.assertTrue(sfx_instance5.played)

        # Stop all sounds playing on the sfx track to start the next test
        track_sfx.stop_all_sounds()
        self.advance_real_time()
        self.mc.bcp_processor.send.reset_mock()
        self.assertEqual(track_sfx.get_sound_players_in_use_count(), 0)
        self.assertEqual(track_sfx.get_sound_queue_count(), 0)

        # Test simultaneous_limit in sound group (skip stealing method)
        self.assertIn('drum_group', self.mc.sounds)
        drum_group = self.mc.sounds['drum_group']
        self.assertEqual(drum_group.simultaneous_limit, 3)
        if SoundStealingMethod is not None:
            self.assertEqual(drum_group.stealing_method, SoundStealingMethod.skip)

        drum_group_instance1 = drum_group.play(settings={'events_when_played': ['drum_group_instance1_played']})
        drum_group_instance2 = drum_group.play(settings={'events_when_played': ['drum_group_instance2_played']})
        drum_group_instance3 = drum_group.play(settings={'events_when_played': ['drum_group_instance3_played']})
        drum_group_instance4 = drum_group.play(settings={'events_when_played': ['drum_group_instance4_played']})
        drum_group_instance5 = drum_group.play(settings={'events_when_played': ['drum_group_instance5_played']})
        self.advance_real_time()

        self.mc.bcp_processor.send.assert_any_call('trigger', sound_instance=ANY, name='drum_group_instance1_played')
        self.mc.bcp_processor.send.assert_any_call('trigger', sound_instance=ANY, name='drum_group_instance2_played')
        self.mc.bcp_processor.send.assert_any_call('trigger', sound_instance=ANY, name='drum_group_instance3_played')
        with self.assertRaises(AssertionError):
            self.mc.bcp_processor.send.assert_any_call('trigger', sound_instance=ANY, name='drum_group_instance4_played')
            self.mc.bcp_processor.send.assert_any_call('trigger', sound_instance=ANY, name='drum_group_instance5_played')

        self.assertTrue(drum_group_instance1.played)
        self.assertTrue(drum_group_instance2.played)
        self.assertTrue(drum_group_instance3.played)
        self.assertIsNone(drum_group_instance4)
        self.assertIsNone(drum_group_instance5)

    def test_sound_player_parameters(self):
        """ Tests sound parameters overridden in the sound_player"""

        if SoundSystem is None or self.mc.sound_system is None:
            log = logging.getLogger('TestAudio')
            log.warning("Sound system is not enabled - skipping audio tests")
            self.skipTest("Sound system is not enabled")

        self.assertIsNotNone(self.mc.sound_system)
        interface = self.mc.sound_system.audio_interface
        if interface is None:
            log = logging.getLogger('TestAudio')
            log.warning("Sound system audio interface could not be loaded - skipping audio tests")
            self.skipTest("Sound system audio interface could not be loaded")

        self.assertIsNotNone(interface)

        # Mock BCP send method
        self.mc.bcp_processor.send = MagicMock()
        self.mc.bcp_processor.enabled = True

        track_sfx = interface.get_track_by_name("sfx")
        self.assertIsNotNone(track_sfx)
        self.assertEqual(track_sfx.name, "sfx")
        self.assertEqual(track_sfx.max_simultaneous_sounds, 8)

        self.advance_real_time(1)

        self.assertIn('264828_text', self.mc.sounds)  # .wav
        text_sound = self.mc.sounds['264828_text']

        # Test sound played from sound_player with all default parameter values
        self.mc.events.post('play_sound_text_default_params')
        self.advance_real_time()
        self.mc.bcp_processor.send.assert_any_call('trigger', sound_instance=ANY, name='text_sound_played')
        status = track_sfx.get_status()
        self.assertEqual(status[0]['sound_id'], text_sound.id)
        instance_id = status[0]['sound_instance_id']
        text_sound_instance = track_sfx.get_playing_sound_instance_by_id(instance_id)
        self.assertIsNotNone(text_sound_instance)
        self.assertEqual(text_sound_instance.volume, 0.5)
        self.assertEqual(text_sound_instance.loops, 7)
        self.assertEqual(text_sound_instance.priority, 0)
        self.assertEqual(text_sound_instance.start_at, 0)
        self.assertEqual(text_sound_instance.fade_out, 0)
        self.assertEqual(text_sound_instance.fade_out, 0)
        self.assertIsNone(text_sound_instance.max_queue_time)
        self.assertEqual(text_sound_instance.simultaneous_limit, 3)

        track_sfx.stop_all_sounds()
        self.advance_real_time()
        self.mc.bcp_processor.send.assert_any_call('trigger', sound_instance=ANY, name='text_sound_stopped')

        # Now test sound played from sound_player with overridden parameter values
        self.mc.events.post('play_sound_text_param_set_1')
        self.advance_real_time()
        self.mc.bcp_processor.send.assert_any_call('trigger', sound_instance=ANY, name='text_sound_played_param_set_1')
        status = track_sfx.get_status()
        self.assertEqual(status[0]['sound_id'], text_sound.id)
        instance_id = status[0]['sound_instance_id']
        text_sound_instance = track_sfx.get_playing_sound_instance_by_id(instance_id)
        self.assertIsNotNone(text_sound_instance)
        self.assertEqual(text_sound_instance.volume, 0.67)
        self.assertEqual(text_sound_instance.loops, 2)
        self.assertEqual(text_sound_instance.priority, 1000)
        self.assertEqual(text_sound_instance.start_at, 0.05)
        self.assertEqual(text_sound_instance.fade_in, 0.25)
        self.assertEqual(text_sound_instance.fade_out, 0.1)
        self.assertEqual(text_sound_instance.max_queue_time, 0.15)
        self.assertEqual(text_sound_instance.simultaneous_limit, 3)

        track_sfx.stop_all_sounds()
        self.advance_real_time()
        self.mc.bcp_processor.send.assert_any_call('trigger', sound_instance=ANY, name='text_sound_stopped_param_set_1')

    def test_track_player(self):
        """Tests the track_player"""

        if SoundSystem is None or self.mc.sound_system is None:
            log = logging.getLogger('TestAudio')
            log.warning("Sound system is not enabled - skipping audio tests")
            self.skipTest("Sound system is not enabled")

        # Mock BCP send method
        self.mc.bcp_processor.send = MagicMock()
        self.mc.bcp_processor.enabled = True

        self.assertIsNotNone(self.mc.sound_system)
        interface = self.mc.sound_system.audio_interface
        if interface is None:
            log = logging.getLogger('TestAudio')
            log.warning("Sound system audio interface could not be loaded - skipping audio tests")
            self.skipTest("Sound system audio interface could not be loaded")

        self.assertIsNotNone(interface)

        track_music = interface.get_track_by_name("music")
        self.assertIsNotNone(track_music)
        self.assertEqual(track_music.name, "music")
        self.assertEqual(track_music.max_simultaneous_sounds, 1)

        self.advance_real_time(1)

        self.assertIn('263774_music', self.mc.sounds)  # .wav
        sound_music = self.mc.sounds['263774_music']
        self.assertFalse(sound_music.loaded)

        self.mc.events.post('load_music')
        self.advance_real_time(2)
        self.assertTrue(sound_music.loaded)

        # TODO: Improve test (need some automated status checks for track control)
        instance = sound_music.play()
        self.advance_real_time(2)
        self.mc.events.post('pause_music_track')
        self.advance_real_time(2)
        self.mc.bcp_processor.send.assert_any_call('trigger', track='music', name='music_track_paused')
        self.mc.bcp_processor.send.reset_mock()

        self.mc.events.post('resume_music_track')
        self.advance_real_time(2)
        self.mc.bcp_processor.send.assert_any_call('trigger', track='music', name='music_track_played')
        self.mc.bcp_processor.send.assert_any_call('trigger', track='music', name='keep_going')
        self.mc.bcp_processor.send.reset_mock()

        self.mc.events.post('stop_all_tracks')
        self.advance_real_time(2)
        self.mc.bcp_processor.send.assert_any_call('trigger', track='music', name='music_track_stopped')
        self.mc.bcp_processor.send.reset_mock()

        self.mc.events.post('play_music_track')
        self.advance_real_time(1)
        self.mc.bcp_processor.send.assert_any_call('trigger', track='music', name='music_track_played')
        self.mc.bcp_processor.send.reset_mock()

        instance = self.mc.sounds['263774_music'].play()
        self.advance_real_time(2)
        self.mc.events.post('set_music_track_volume_quiet')
        self.advance_real_time(2)
        self.mc.events.post('set_music_track_volume_loud')
        self.advance_real_time(2)
        self.mc.events.post('stop_all_sounds_on_music_track')
        self.advance_real_time(1)

        self.assertTrue(sound_music.loaded)
        self.mc.events.post('unload_music')
        self.advance_real_time()
        self.assertFalse(sound_music.loaded)

        def test_sound_looping(self):
            """ Tests the sound looping features """

            if SoundSystem is None or self.mc.sound_system is None:
                log = logging.getLogger('TestAudio')
                log.warning("Sound system is not enabled - skipping audio tests")
                self.skipTest("Sound system is not enabled")

            self.assertIsNotNone(self.mc.sound_system)
            interface = self.mc.sound_system.audio_interface
            if interface is None:
                log = logging.getLogger('TestAudio')
                log.warning("Sound system audio interface could not be loaded - skipping audio tests")
                self.skipTest("Sound system audio interface could not be loaded")

            self.assertIsNotNone(interface)

            # Mock BCP send method
            self.mc.bcp_processor.send = MagicMock()
            self.mc.bcp_processor.enabled = True

            track_music = interface.get_track_by_name("music")
            self.assertIsNotNone(track_music)
            self.assertEqual(track_music.name, "music")
            self.assertEqual(track_music.max_simultaneous_sounds, 1)

            # Because these tests can be run at different audio settings depending upon hardware,
            # sample lengths and positions in bytes are not reliable to compare for testing. This
            # test will convert sample lengths/positions back to time for comparisons based on the
            # audio interface settings actually used in the test.
            settings = self.mc.sound_system.audio_interface.get_settings()
            self.assertIsNotNone(settings)
            seconds_to_bytes_factor = settings['seconds_to_bytes_factor']

            self.advance_real_time(1)

            self.assertIn('looptest', self.mc.sounds)  # .wav
            sound_looptest = self.mc.sounds['looptest']
            self.assertFalse(sound_looptest.loaded)
            self.assertAlmostEqual(sound_looptest.loop_start_at, 1.8461538, 4)
            self.assertAlmostEqual(sound_looptest.loop_end_at, 3.6923077, 4)
            self.assertEqual(sound_looptest.loops, 3)

            instance1 = track_music.play_sound(sound_looptest, context=None, settings={'volume': 1.0})
            self.advance_real_time()

            status = track_music.get_status()
            self.assertEqual(status[0]['sound_instance_id'], instance1.id)
            self.assertEqual(status[0]['sound_id'], sound_looptest.id)
            self.assertEqual(status[0]['status'], "playing")
            self.assertEqual(status[0]['current_loop'], 0)
            self.assertEqual(status[0]['loops'], 3)
            self.assertAlmostEqual(status[0]['loop_start_pos'] / seconds_to_bytes_factor, 1.8461538, 4)
            self.assertAlmostEqual(status[0]['loop_end_pos'] / seconds_to_bytes_factor, 3.6923077, 4)
            self.mc.bcp_processor.send.assert_any_call('trigger', sound_instance=ANY, name='looptest_played')
            self.mc.bcp_processor.send.reset_mock()

            self.advance_real_time(10)
            instance1.stop()
            self.advance_real_time()

            status = track_music.get_status()
            self.assertEqual(status[0]['sound_instance_id'], instance1.id)
            self.assertEqual(status[0]['current_loop'], 3)
            self.assertEqual(status[0]['loops'], 0)
            self.assertAlmostEqual(status[0]['loop_start_pos'] / seconds_to_bytes_factor, 1.8461538, 4)
            self.assertAlmostEqual(status[0]['loop_end_pos'] / seconds_to_bytes_factor, 5.8846259, 4)

            self.mc.bcp_processor.send.assert_any_call('trigger', sound_instance=ANY, name='looptest_looping')
            self.mc.bcp_processor.send.assert_any_call('trigger', sound_instance=ANY, name='looptest_stopped')

            # Now start playing the song after the end of the loop end (should loop back to the beginning of the loop)
            instance1 = track_music.play_sound(sound_looptest, context=None,
                                               settings={'volume': 1.0, 'start_at': 3.8, 'loops': 5})
            self.advance_real_time()

            status = track_music.get_status()
            self.assertEqual(status[0]['sound_instance_id'], instance1.id)
            self.assertEqual(status[0]['sound_id'], sound_looptest.id)
            self.assertEqual(status[0]['status'], "playing")
            self.assertEqual(status[0]['current_loop'], 0)
            self.assertEqual(status[0]['loops'], 5)
            self.assertGreater(status[0]['sample_pos'] / seconds_to_bytes_factor, 3.8)
            self.assertAlmostEqual(status[0]['loop_start_pos'] / seconds_to_bytes_factor, 1.8461538, 4)
            self.assertAlmostEqual(status[0]['loop_end_pos'] / seconds_to_bytes_factor, 3.6923077, 4)
            self.mc.bcp_processor.send.assert_any_call('trigger', sound_instance=ANY, name='looptest_played')
            self.mc.bcp_processor.send.reset_mock()

            self.advance_real_time(2.5)

            status = track_music.get_status()
            self.assertEqual(status[0]['sound_instance_id'], instance1.id)
            self.assertEqual(status[0]['current_loop'], 1)
            self.assertEqual(status[0]['loops'], 4)
            self.assertGreater(status[0]['sample_pos'] / seconds_to_bytes_factor, 1.8461538)
            self.assertLess(status[0]['sample_pos'] / seconds_to_bytes_factor, 5.8846259)
            self.assertAlmostEqual(status[0]['loop_start_pos'] / seconds_to_bytes_factor, 1.8461538, 4)
            self.assertAlmostEqual(status[0]['loop_end_pos'] / seconds_to_bytes_factor, 3.6923077, 4)

            instance1.stop()
            self.advance_real_time()

            self.mc.bcp_processor.send.assert_any_call('trigger', sound_instance=ANY, name='looptest_looping')
            self.mc.bcp_processor.send.assert_any_call('trigger', sound_instance=ANY, name='looptest_stopped')
            self.mc.bcp_processor.send.reset_mock()

            # Now test a voice-stealing scenario to ensure sound settings are copied correctly
            track_voice = interface.get_track_by_name("voice")
            self.assertIsNotNone(track_voice)
            self.assertEqual(track_voice.name, "voice")
            self.assertEqual(track_voice.max_simultaneous_sounds, 1)

            self.assertIn('104457_moron_test', self.mc.sounds)
            sound_moron_test = self.mc.sounds['104457_moron_test']

            instance2 = track_voice.play_sound(sound_moron_test, context=None,
                                               settings={'volume': 1.0, 'loops': 0, 'priority': 900})
            self.advance_real_time()

            status = track_voice.get_status()
            self.assertEqual(status[0]['sound_instance_id'], instance2.id)
            self.assertEqual(status[0]['sound_id'], sound_moron_test.id)
            self.assertEqual(status[0]['status'], "playing")
            self.assertEqual(status[0]['current_loop'], 0)
            self.assertEqual(status[0]['loops'], 0)
            self.assertAlmostEqual(status[0]['loop_start_pos'] / seconds_to_bytes_factor, 0.0, 3)
            self.assertAlmostEqual(status[0]['loop_end_pos'] / seconds_to_bytes_factor, 7.39048, 3)
            self.mc.bcp_processor.send.assert_any_call('trigger', sound_instance=ANY, name='moron_test_played')
            self.mc.bcp_processor.send.reset_mock()

            self.advance_real_time(1.0)

            self.assertIn('113690_test', self.mc.sounds)
            sound_test = self.mc.sounds['113690_test']

            # Play sound with higher priority than the one currently playing to make sure the sound replacing
            # code copies all the sound settings to the new sound (this test specifically addresses a bug
            # introduced after adding loop point settings).
            instance3 = track_voice.play_sound(sound_test, context=None,
                                               settings={'volume': 1.0, 'loops': 0, 'priority': 1000})
            self.advance_real_time(0.2)

            status = track_voice.get_status()
            self.assertEqual(status[0]['sound_instance_id'], instance3.id)
            self.assertEqual(status[0]['sound_id'], sound_test.id)
            self.assertEqual(status[0]['status'], "playing")
            self.assertEqual(status[0]['current_loop'], 0)
            self.assertEqual(status[0]['loops'], 0)
            self.assertAlmostEqual(status[0]['loop_start_pos'] / seconds_to_bytes_factor, 0.0, 3)
            self.assertAlmostEqual(status[0]['loop_end_pos'] / seconds_to_bytes_factor, 6.95655, 3)

        # TODO: Add integration test for sound_player
        # TODO: Add integration test for track_player

        """
        # Add another track with the same name (should not be allowed)
        # Add another track with the same name, but different casing (should not be allowed)
        # Attempt to create track with max_simultaneous_sounds > 32 (the current max)
        # Attempt to create track with max_simultaneous_sounds < 1 (the current max)
        # Add up to the maximum number of tracks allowed
        # There should now be the maximum number of tracks allowed
        # Try to add another track (more than the maximum allowed)

        # TODO: Tests to write:
        # Load sounds (wav, ogg, flac, unsupported format)
        # Play a sound
        # Play two sounds on track with max_simultaneous_sounds = 1 (test sound queue,
        time expiration, priority scenarios)
        # Play a sound on each track simultaneously
        # Stop all sounds on track
        # Stop all sounds on all tracks
        # Ducking
        # Configuration file tests (audio interface, tracks, sounds, sound player, sound
        # trigger events, etc.)
        #
        """



"""Widget showing a line."""
from typing import Optional

from kivy.graphics import Line as KivyLine
from kivy.graphics.context_instructions import Color, Scale, Rotate
from kivy.properties import (ListProperty, NumericProperty, OptionProperty,
                             BooleanProperty)

from mpfmc.uix.widget import Widget
from mpfmc.core.utils import center_of_points_list

MYPY = False
if MYPY:   # pragma: no cover
    from mpfmc.core.mc import MpfMc


class Line(Widget):

    """Widget showing a line."""

    widget_type_name = 'Line'
    animation_properties = ('color', 'thickness', 'opacity', 'points', 'rotation', 'scale')

    def __init__(self, mc: "MpfMc", config: dict, key: Optional[str] = None, **kwargs) -> None:
        del kwargs
        super().__init__(mc=mc, config=config, key=key)

        # The points in this widget are always relative to the bottom left corner
        self.anchor_pos = ("left", "bottom")

        # Bind to all properties that when changed need to force
        # the widget to be redrawn
        self.bind(color=self._draw_widget,
                  points=self._draw_widget,
                  thickness=self._draw_widget,
                  rotation=self._draw_widget,
                  scale=self._draw_widget)

        self._draw_widget()

    def _draw_widget(self, *args) -> None:
        """Establish the drawing instructions for the widget."""
        del args

        if self.canvas is None:
            return

        # TODO: allow user to set rotation/scale origin
        center = center_of_points_list(self.points)
        self.canvas.clear()

        with self.canvas:
            Color(*self.color)
            Scale(self.scale, origin=center)
            Rotate(angle=self.rotation, origin=center)
            KivyLine(points=self.points,
                     width=self.thickness,
                     cap=self.cap,
                     joint=self.joint,
                     cap_precision=self.cap_precision,
                     joint_precision=self.joint_precision,
                     close=self.close)

    #
    # Properties
    #

    points = ListProperty()
    '''The list of points to use to draw the widget in (x1, y1, x2, y2...)
    format.

    :attr:`points` is a :class:`~kivy.properties.ListProperty`.
    '''

    cap = OptionProperty("round", options=["none", "square", "round"])
    '''The cap of the line, defaults to 'round'. Can be one of 'none',
    'square' or 'round'
    '''

    cap_precision = NumericProperty(10)
    '''Number of iterations for drawing the "round" cap, defaults to 10. The
    cap_precision must be at least 1.
    '''

    close = BooleanProperty(False)
    '''If True, the line will be closed.
    '''

    joint = OptionProperty("round", options=["none", "round", "bevel", "miter"])
    '''The join of the line, defaults to 'round'. Can be one of 'none', 'round',
    'bevel', 'miter'.
    '''

    joint_precision = NumericProperty(10)
    '''Number of iterations for drawing the "round" joint, defaults to 10. The
    joint_precision must be at least 1.
    '''

    thickness = NumericProperty(1.0)
    '''Width of the line.

    :attr:`thickness` is a :class:`~kivy.properties.NumericProperty` and defaults
    to 1.0.
    '''

    rotation = NumericProperty(0)
    '''Rotation angle value of the widget.

    :attr:`rotation` is an :class:`~kivy.properties.NumericProperty` and defaults to
    0.
    '''

    scale = NumericProperty(1.0)
    '''Scale value of the widget.

    :attr:`scale` is an :class:`~kivy.properties.NumericProperty` and defaults to
    1.0.
    '''


widget_classes = [Line]



cdef extern from *:
    ctypedef bint bool

# ---------------------------------------------------------------------------
#    GStreamer declarations from gst.h
# ---------------------------------------------------------------------------
cdef extern from 'gst/gst.h':
    ctypedef void *GstPipeline
    ctypedef void *GstElement
    ctypedef void *GstBus
    ctypedef void *GstPad
    ctypedef void *GstSample
    ctypedef void *GstBuffer
    ctypedef void *GstMemory
    ctypedef void *GstBin
    ctypedef void *GstCaps
    ctypedef void *GstTimedValueControlSource
    ctypedef void *GstQuery
    ctypedef void *GObject
    ctypedef void (*appcallback_t)(void *, int, int, char *, int)
    ctypedef void (*buscallback_t)(void *, GstMessage *)
    ctypedef unsigned int guint
    ctypedef unsigned long gulong
    ctypedef unsigned char guint8
    ctypedef unsigned int guint32
    ctypedef unsigned int gsize
    ctypedef void *gpointer
    ctypedef const void *gconstpointer
    ctypedef char gchar
    ctypedef char const_gchar 'const gchar'
    ctypedef int gint
    ctypedef long int gint64
    ctypedef unsigned long long GstClockTime
    ctypedef int gboolean

    ctypedef enum GstState:
        GST_STATE_VOID_PENDING
        GST_STATE_NULL
        GST_STATE_READY
        GST_STATE_PAUSED
        GST_STATE_PLAYING

    ctypedef enum GstFormat:
        GST_FORMAT_DEFAULT
        GST_FORMAT_BYTES
        GST_FORMAT_TIME

    ctypedef enum GstSeekFlags:
        GST_SEEK_FLAG_KEY_UNIT
        GST_SEEK_FLAG_FLUSH

    ctypedef enum GstStateChangeReturn:
        pass

    ctypedef struct GError:
        int code
        char *message

    ctypedef struct GstAudioInfo:
        gint rate
        gint channels
        gint bpf

    ctypedef struct GstMapInfo:
        GstMemory *memory
        GstMapFlags flags
        guint8 *data
        gsize size
        gsize maxsize

    ctypedef enum GstMapFlags:
        GST_MAP_READ
        GST_MAP_WRITE
        GST_MAP_FLAG_LAST

    ctypedef enum GstMessageType:
        GST_MESSAGE_EOS
        GST_MESSAGE_ERROR
        GST_MESSAGE_WARNING
        GST_MESSAGE_INFO

    ctypedef struct GstMessage:
        GstMessageType type

    int GST_SECOND
    bool gst_init_check(int *argc, char ***argv, GError **error)
    bool gst_is_initialized()
    void gst_deinit()
    void gst_version(guint *major, guint *minor, guint *micro, guint *nano)
    GstElement *gst_parse_launch(const_gchar *pipeline_description, GError **error)
    GstElement *gst_element_factory_make(const_gchar *factoryname, const_gchar *name)
    GstElement *gst_bin_new(const_gchar *name)
    bool gst_bin_add(GstBin *bin, GstElement *element)
    bool gst_bin_remove(GstBin *bin, GstElement *element)
    GstElement *gst_bin_get_by_name(GstBin *bin, const_gchar *name)
    void gst_object_unref(void *pointer) nogil
    bool gst_element_query(GstElement *element, GstQuery *query)
    GstQuery *gst_query_new_duration(GstFormat format)
    void gst_query_parse_duration(GstQuery *query, GstFormat *format, gint64 *duration)
    void gst_query_unref(GstQuery *q)
    GstElement *gst_pipeline_new(const_gchar *name)
    GstPad *gst_element_get_static_pad(GstElement *element, const_gchar *name)
    GstPad *gst_element_get_request_pad(GstElement *element, const_gchar *name)
    void gst_element_release_request_pad(GstElement *element, GstPad *pad)
    GstPad *gst_ghost_pad_new(const_gchar *name, GstPad *target)
    bool gst_element_add_pad(GstElement *element, GstPad *pad)
    void gst_bus_enable_sync_message_emission(GstBus *bus)
    GstBus *gst_pipeline_get_bus(GstPipeline *pipeline)
    GstBuffer *gst_sample_get_buffer(GstSample *sample) nogil
    void gst_buffer_unref(GstBuffer *buffer) nogil
    void gst_sample_unref(GstSample *sample) nogil
    gboolean gst_buffer_map(GstBuffer *buffer, GstMapInfo *info, GstMapFlags flags) nogil
    void gst_buffer_unmap(GstBuffer *buffer, GstMapInfo *info) nogil
    GstCaps *gst_sample_get_caps(GstSample *sample)
    GstStateChangeReturn gst_element_get_state(
            GstElement *element, GstState *state, GstState *pending,
            GstClockTime timeout) nogil
    GstStateChangeReturn gst_element_set_state(
            GstElement *element, GstState state) nogil
    void g_signal_emit_by_name(gpointer instance, const_gchar *detailed_signal,
            void *retvalue)
    GstCaps *gst_caps_from_string(const_gchar *string)
    void gst_caps_unref(GstCaps *caps)
    bool gst_audio_info_from_caps(GstAudioInfo *info, const GstCaps *caps)
    void g_error_free(GError *error)
    bool gst_element_link(GstElement *src, GstElement *dest) nogil
    bool gst_element_link_filtered(GstElement *src, GstElement *dest, GstCaps *filter) nogil
    bool gst_element_query_position(
            GstElement *element, GstFormat format, gint64 *cur) nogil
    bool gst_element_query_duration(
            GstElement *element, GstFormat format, gint64 *cur) nogil
    bool gst_element_seek_simple(
            GstElement *element, GstFormat format,
            GstSeekFlags seek_flags, gint64 seek_pos) nogil
    void gst_message_parse_error(
            GstMessage *message, GError **gerror, char **debug)
    void gst_message_parse_warning(
            GstMessage *message, GError **gerror, char **debug)
    void gst_message_parse_info(
            GstMessage *message, GError **gerror, char **debug)


# ---------------------------------------------------------------------------
#    glib declarations from glib.h
# ---------------------------------------------------------------------------
cdef extern from 'glib.h':
    int glib_major_version
    int glib_minor_version
    int glib_micro_version

    ctypedef void (*GFunc)(gpointer data, gpointer user_data)

    # Memory management
    ctypedef void (*GDestroyNotify)(gpointer data)
    ctypedef gpointer (*GReallocFunc)(gpointer data, gsize size)
    void g_free(gpointer mem)
    gpointer g_malloc(gsize n_bytes)
    gpointer g_realloc(gpointer mem, gsize n_bytes)

    # Linked list
    ctypedef _GSList GSList
    ctypedef struct _GSList:
        gpointer data
        GSList *next
    GSList* g_slist_append(GSList *list, gpointer data) nogil
    GSList* g_slist_prepend(GSList *list, gpointer data) nogil
    GSList* g_slist_remove(GSList *list, gconstpointer data) nogil
    void g_slist_free (GSList *list) nogil
    guint g_slist_length (GSList *list) nogil
    GSList *g_slist_reverse(GSList *list) nogil
    void g_slist_foreach(GSList *list, GFunc func, gpointer user_data) nogil
    gpointer g_slist_nth_data(GSList *list, guint n) nogil

    # Memory slices
    gpointer g_slice_alloc(gsize block_size) nogil
    gpointer g_slice_alloc0(gsize block_size) nogil
    void g_slice_free1(gsize block_size, gpointer mem_block) nogil
    gpointer g_slice_copy(gsize block_size, gconstpointer mem_block) nogil

    # Array
    ctypedef struct GArray:
        gchar *data
        guint len
    GArray *g_array_new(gboolean zero_terminated, gboolean clear_, guint element_size) nogil
    GArray *g_array_sized_new(gboolean zero_terminated, gboolean clear_, guint element_size, guint reserved_size) nogil
    GArray *g_array_set_size(GArray *array, guint length) nogil
    gchar *g_array_free(GArray *array, gboolean free_segment) nogil

# ---------------------------------------------------------------------------
#    GStreamer helper functions defined in gstreamer_helper.h
# ---------------------------------------------------------------------------
cdef extern from 'gstreamer_helper.h':
    void g_gst_log_error(const_gchar *file, const_gchar *function, gint line, GObject *object, const_gchar *message) nogil
    void g_gst_log_warning(const_gchar *file, const_gchar *function, gint line, GObject *object, const_gchar *message) nogil
    void g_gst_log_info(const_gchar *file, const_gchar *function, gint line, GObject *object, const_gchar *message) nogil
    void g_gst_log_debug(const_gchar *file, const_gchar *function, gint line, GObject *object, const_gchar *message) nogil
    gboolean g_object_get_bool(GstElement *element, char *name) nogil
    GstSample *c_appsink_pull_preroll(GstElement *appsink) nogil
    GstSample *c_appsink_pull_sample(GstElement *appsink) nogil
    gulong c_bus_connect_message(GstBus *bus,
            buscallback_t callback, void *userdata)
    void c_signal_disconnect(GstElement *appsink, gulong handler_id)

    void g_array_insert_val_uint(GArray *array, guint index, guint value) nogil
    void g_array_insert_val_uint8(GArray *array, guint index, guint8 value) nogil
    guint g_array_index_uint(GArray* array, guint index) nogil
    guint8 g_array_index_uint8(GArray* array, guint index) nogil
    void g_array_set_val_uint(GArray *array, guint index, guint value) nogil
    void g_array_set_val_uint8(GArray *array, guint index, guint8 value) nogil



from typing import List

from kivy.properties import NumericProperty, ListProperty

from mpfmc.effects.color_dmd import ColorDmdEffect
from mpfmc.effects.dot_filter import DotFilterEffect
from mpfmc.effects.monochrome import MonochromeEffect
from mpfmc.effects.reduce import ReduceEffect
from mpfmc.effects.colorize import ColorizeEffect
from mpfmc.effects.gain import GainEffect

MYPY = False
if MYPY:   # pragma: no cover
    from kivy.uix.effectwidget import EffectBase


class DmdEffect(ColorDmdEffect):

    """GLSL effect to render an on-screen DMD to look like individual round pixels."""

    luminosity = ListProperty([.299, .587, .114])
    '''This defines the luminosity factor for each color channel. The value
    for each channel must be between 0.0 and 1.0.

    :attr:`luminosity` is a :class:`ListProperty` defaults to
    (.299, .587, .114)
    '''

    dot_color = ListProperty([1, 0.4, 0, 0])
    '''This defines the dot color to be used in the effect.

    :attr:`dot_color` is a :class:`ListProperty` defaults to
    (1, 0.4, 1, 0)
    '''

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def get_effects(self) -> List["EffectBase"]:
        effects = []

        if bool(self.dot_filter):
            effects.append(DotFilterEffect(
                width=self.width,
                height=self.height,
                dots_x=self.dots_x,
                dots_y=self.dots_y,
                blur=self.blur,
                dot_size=self.dot_size,
                background_color=self.background_color
            ))

        effects.append(MonochromeEffect(luminosity=self.luminosity))

        if self.shades > 0:
            effects.append(ReduceEffect(shades=self.shades))

        effects.append(ColorizeEffect(tint_color=self.dot_color))
        effects.append(GainEffect(gain=self.gain))

        return effects


effect_cls = DmdEffect
name = 'dmd'



#!python
#cython: embedsignature=True, language_level=3
"""
Audio Library

This library requires the SDL2, SDL_Mixer, and Gstreamer libraries.
"""


from libc.stdio cimport FILE, fopen, fprintf, sprintf, fflush
from libc.string cimport memset, memcpy
from cpython.mem cimport PyMem_Malloc, PyMem_Realloc, PyMem_Free
cimport cpython.pycapsule as pycapsule

from math import pow
import logging

from mpfmc.core.audio.sdl2 cimport *
from mpfmc.core.audio.gstreamer cimport *
from mpfmc.core.audio.track cimport *
from mpfmc.core.audio.track_standard cimport *
from mpfmc.core.audio.track_sound_loop cimport *
from mpfmc.core.audio.sound_file cimport *
from mpfmc.core.audio.audio_exception import AudioException
from mpfmc.core.audio.playlist_controller import PlaylistController


# ---------------------------------------------------------------------------
#    Various audio engine setting values
# ---------------------------------------------------------------------------
DEF MAX_TRACKS = 8

# The maximum number of markers that can be specified for a single sound
DEF MAX_MARKERS = 16

# The number of seconds over which to perform a quick fade to avoid pops and
# clicks
DEF QUICK_FADE_DURATION_SECS = 0.05


# ---------------------------------------------------------------------------
#    Global GStreamer helper functions
# ---------------------------------------------------------------------------
def _gst_init():
    """Initializes the GStreamer library"""
    if gst_is_initialized():
        return True
    cdef int argc = 0
    cdef char **argv = NULL
    cdef GError *error
    if not gst_init_check(&argc, &argv, &error):
        msg = 'Unable to initialize GStreamer: code={} message={}'.format(
                error.code, <bytes>error.message)
        raise AudioException(msg)

def get_gst_version():
    """Returns the current version of GStreamer"""
    cdef unsigned int major, minor, micro, nano
    gst_version(&major, &minor, &micro, &nano)
    return major, minor, micro, nano


# ---------------------------------------------------------------------------
#    AudioInterface class
# ---------------------------------------------------------------------------
cdef class AudioInterface:
    """
    The AudioInterface class provides a management wrapper around the SDL2 and SDL_Mixer
    libraries.
    """
    cdef list tracks
    cdef dict playlist_controllers
    cdef object mc
    cdef object log

    cdef AudioCallbackData audio_callback_data

    def __cinit__(self, *args, **kw):
        pass

    def __init__(self, rate=44100, channels=2, buffer_samples=4096):
        """
        Initializes the AudioInterface.
        Args:
            rate: The audio sample rate used in the library
            channels: The number of channels to use (1=mono, 2=stereo)
            buffer_samples: The audio buffer size to use (in number of samples, must be power of two)
        """
        self.log = logging.getLogger("AudioInterface")

        # Initialize threading in the extension library and acquire the Python global interpreter lock
        PyEval_InitThreads()

        if channels not in (1, 2):
            self.log.error('Channels is required to be either 1 (mono) or 2 (stereo)')
            raise AudioException("Unable to initialize Audio Interface: "
                                 "Channels is required to be either 1 (mono) or 2 (stereo)")

        # Make sure buffer samples is a power of two (required by SDL2)
        if not AudioInterface.power_of_two(buffer_samples):
            self.log.error('Buffer samples is required to be a power of two')
            raise AudioException("Unable to initialize Audio Interface: "
                                 "Buffer samples is required to be a power of two")

        # Warn if a small buffer is used
        if buffer_samples <= 1024:
            self.log.warning('NOTE: You may experience noise and other undesirable sound artifacts '
                             'when you set your buffer at 1024 or smaller.')

        # Initialize the SDL audio system
        if SDL_InitSubSystem(SDL_INIT_AUDIO) < 0:
            self.log.error('SDL_InitSubSystem error - %s' % SDL_GetError())
            raise AudioException('Unable to initialize SDL (SDL_InitSubSystem call failed: %s)' % SDL_GetError())

        # Initialize the SDL_Mixer library to establish the output audio format and encoding
        # (sample rate, bit depth, buffer size)
        if Mix_OpenAudio(rate, AUDIO_S16SYS, channels, buffer_samples):
            self.log.error('Mix_OpenAudio error - %s' % SDL_GetError())
            raise AudioException('Unable to open audio for output (Mix_OpenAudio failed: %s)' % SDL_GetError())

        # We want to use as little resources as possible for SDL_Mixer as we will just be using the custom
        # music player hook to play audio (no mixer channels needed).
        Mix_AllocateChannels(0)

        # Initialize GStreamer
        self._initialize_gstreamer()

        self.log.info("Initialized")
        self.log.debug("Loaded %s", AudioInterface.get_sdl_version())
        self.log.debug("Loaded %s", AudioInterface.get_sdl_mixer_version())
        self.log.debug("Loaded %s", AudioInterface.get_gstreamer_version())
        self.log.debug("Loaded %s", AudioInterface.get_glib_version())

        # Lock SDL from calling the audio callback functions while we set things up
        SDL_LockAudio()

        # Determine the actual audio format in use by the opened audio device.  This may or may not match
        # the parameters used to initialize the audio interface.
        self.audio_callback_data.buffer_samples = buffer_samples
        Mix_QuerySpec(&self.audio_callback_data.sample_rate,
                      &self.audio_callback_data.format,
                      &self.audio_callback_data.channels)

        # Ensure system is little endian (big endian not supported)
        if not SDL_AUDIO_ISLITTLEENDIAN(self.audio_callback_data.format):
            self.log.error("The audio interface only supports little endian systems in this release. "
                           "Audio features will not be available.")
            raise AudioException("The audio interface only supports little endian systems in this release.")

        # The requested values used to initialize the audio interface.  A pointer to the audio_callback_data
        # structure is passed to the SDL audio callback function and is the source of all audio state
        # and mixing data needed to generate the output signal.
        self.audio_callback_data.bytes_per_sample = SDL_AUDIO_BITSIZE(self.audio_callback_data.format) // 8
        self.audio_callback_data.buffer_size = self.audio_callback_data.buffer_samples * self.audio_callback_data.bytes_per_sample * self.audio_callback_data.channels
        self.audio_callback_data.bytes_per_control_point = self.audio_callback_data.buffer_size // CONTROL_POINTS_PER_BUFFER
        self.audio_callback_data.seconds_to_bytes_factor = self.audio_callback_data.sample_rate * self.audio_callback_data.channels * self.audio_callback_data.bytes_per_sample
        self.audio_callback_data.master_volume = SDL_MIX_MAXVOLUME // 2
        self.audio_callback_data.quick_fade_steps = (<int>(QUICK_FADE_DURATION_SECS *
                                                     self.audio_callback_data.sample_rate *
                                                     self.audio_callback_data.channels *
                                                     self.audio_callback_data.bytes_per_sample
                                                           )) // self.audio_callback_data.bytes_per_control_point
        self.audio_callback_data.silence = 0
        self.audio_callback_data.track_count = 0
        self.audio_callback_data.tracks = <void**>PyMem_Malloc(MAX_TRACKS * sizeof(TrackState*))
        self.audio_callback_data.c_log_file = NULL
        # self.audio_callback_data.c_log_file = fopen("D:\\Temp\\Dev\\MPFMC_AudioLibrary.log", "wb")
        # fprintf(self.audio_callback_data.c_log_file, "---------------------------------------------------------------------------\r\n")
        # fflush(self.audio_callback_data.c_log_file)

        self.log.debug('Settings requested - rate: %d, channels: %d, buffer: %d samples',
                       rate, channels, buffer_samples)
        self.log.debug('Settings in use - rate: %d, channels: %d, buffer: %d samples (%d bytes @ %d bytes per sample)',
                       self.audio_callback_data.sample_rate, self.audio_callback_data.channels,
                       self.audio_callback_data.buffer_samples, self.audio_callback_data.buffer_size,
                       self.audio_callback_data.bytes_per_sample)

        # Unlock the SDL audio callback functions
        SDL_UnlockAudio()

        self.tracks = list()
        self.playlist_controllers = dict()

    def __del__(self):
        """Shut down the audio interface and clean up allocated memory"""
        self.log.debug("Shutting down and cleaning up allocated memory...")

        # Stop audio processing (will stop all SDL callbacks)
        self.shutdown()

        self.audio_callback_data.track_count = 0
        PyMem_Free(self.audio_callback_data.tracks)
        self.audio_callback_data.tracks = NULL

        # Remove tracks
        self.tracks.clear()

        # SDL and SDL_Mixer no longer needed
        Mix_Quit()
        SDL_Quit()

    def _initialize_gstreamer(self):
        """Initialize the GStreamer library"""
        if gst_is_initialized():
            return True

        cdef int argc = 0
        cdef char **argv = NULL
        cdef GError *error
        if not gst_init_check(&argc, &argv, &error):
            msg = 'Unable to initialize gstreamer: code={} message={}'.format(error.code, <bytes>error.message)
            raise AudioException(msg)

    @staticmethod
    def initialize(int rate=44100, int channels=2, int buffer_samples=4096, **kwargs):
        """
        Initializes and retrieves the audio interface instance.
        Args:
            rate: The audio sample rate used in the library
            channels: The number of channels to use (1=mono, 2=stereo)
            buffer_samples: The audio buffer size in number of samples (must be power of two)

        Returns:
            An AudioInterface object instance.
        """
        # Initialize the audio instance and return it
        audio_interface_instance = AudioInterface(rate=rate,
                                                  channels=channels,
                                                  buffer_samples=buffer_samples,
                                                  **kwargs)
        return audio_interface_instance

    @staticmethod
    def power_of_two(int num):
        """ Returns whether or not the supplied number is a power of 2 """
        return ((num & (num - 1)) == 0) and num != 0

    @staticmethod
    def db_to_gain(float db):
        """Converts a value in decibels (-inf to 0.0) to a gain (0.0 to 1.0)"""
        return pow(10, db / 20.0)

    @staticmethod
    def string_to_gain(gain):
        """Converts a string to a gain value (0.0 to 1.0)"""
        cdef str gain_string = str(gain).upper()

        if gain_string.endswith('DB'):
            gain_string = ''.join(i for i in gain_string if not i.isalpha())
            return min(max(AudioInterface.db_to_gain(float(gain_string)), 0.0), 1.0)

        return min(max(float(gain_string), 0.0), 1.0)

    def convert_seconds_to_samples(self, float seconds):
        """Converts the specified number of seconds into samples (based on current sample rate)"""
        return int(self.audio_callback_data.sample_rate * seconds)

    def convert_seconds_to_buffer_length(self, float seconds):
        """Convert the specified number of seconds into a buffer length (based on current
        sample rate, the number of audio channels, and the number of bytes per sample)."""
        return int(seconds * self.audio_callback_data.sample_rate * self.audio_callback_data.channels * self.audio_callback_data.bytes_per_sample)

    def convert_buffer_length_to_seconds(self, int buffer_length):
        """Convert the specified buffer length into a time in seconds (based on current
        sample rate, the number of audio channels, and the number of bytes per sample)."""
        return round(buffer_length / (self.audio_callback_data.sample_rate * self.audio_callback_data.channels * self.audio_callback_data.bytes_per_sample), 3)

    @staticmethod
    def string_to_secs(time):
        """Decodes a string of real-world time into a float of seconds.
        Example inputs:

        234ms
        2s
        None

        If no "s" or "ms" is provided, this method assumes "seconds."

        If time is 'None' or a string of 'None', this method returns 0.0.

        Returns:
            Float. The examples listed above return 0.234, 2.0 and 0.0,
            respectively
        """
        cdef str time_string = str(time).upper()

        if time_string.endswith('MS') or time_string.endswith('MSEC'):
            time_string = ''.join(i for i in time_string if not i.isalpha())
            return float(time_string) / 1000.0

        elif time_string.endswith('S') or time_string.endswith('SEC'):
            time_string = ''.join(i for i in time_string if not i.isalpha())
            return float(time_string)

        elif 'D' in time_string:
            time_string = ''.join(i for i in time_string if not i.isalpha())
            return float(time_string) * 86400

        elif 'H' in time_string:
            time_string = ''.join(i for i in time_string if not i.isalpha())
            return float(time_string) * 3600

        elif 'M' in time_string:
            time_string = ''.join(i for i in time_string if not i.isalpha())
            return float(time_string) * 60

        elif not time_string or time_string == 'NONE':
            return 0.0

        else:
            try:
                return float(time_string)
            except:
                return 0.0

    def string_to_samples(self, samples):
        """
        Converts a string to an integer value representing the number of samples.
        Args:
            samples: String containing either a time string or an integer number
                of samples.

        Returns:
            Integer containing the number of samples.
        """
        cdef str samples_string = str(samples).upper()

        # If the string contains only digits we assume it is already in samples
        try:
            return int(float(samples_string))
        except:
            pass

        # Time strings are also permitted and will be converted to seconds
        # and then to samples using the current sample rate.
        cdef float seconds = AudioInterface.string_to_secs(samples_string)
        return int(self.audio_callback_data.sample_rate * seconds)

    @classmethod
    def get_sdl_version(cls):
        """
        Returns the version of the SDL library
        :return: SDL library version string
        """
        cdef SDL_version version
        SDL_GetVersion(&version)
        return 'SDL {}.{}.{}'.format(version.major, version.minor, version.patch)

    @classmethod
    def get_sdl_mixer_version(cls):
        """
        Returns the version of the dynamically linked SDL_Mixer library
        :return: SDL_Mixer library version string
        """
        cdef const SDL_version *version = Mix_Linked_Version()
        return 'SDL_Mixer {}.{}.{}'.format(version.major, version.minor, version.patch)

    @classmethod
    def get_gstreamer_version(cls):
        """
        Returns the version of the GStreamer library
        :return: GStreamer library version string
        """
        gst_version = get_gst_version()
        return 'GStreamer {}.{}.{}.{}'.format(
            gst_version[0], gst_version[1], gst_version[2], gst_version[3])

    @classmethod
    def get_glib_version(cls):
        """
        Returns the version of the GLib library
        :return: GLib library version string
        """
        return 'GLib {}.{}.{}'.format(
            glib_major_version, glib_minor_version, glib_micro_version)

    def supported_extensions(self):
        """
        Get the file extensions that are supported by the audio interface.
        Returns:
            A list of file extensions supported.
        """
        return ["wav", "ogg", "flac",]

    def get_master_volume(self):
        return round(self.audio_callback_data.master_volume / SDL_MIX_MAXVOLUME, 2)

    def set_master_volume(self, float volume):
        SDL_LockAudio()
        self.audio_callback_data.master_volume = <Uint8>min(max(volume * SDL_MIX_MAXVOLUME, 0), SDL_MIX_MAXVOLUME)
        SDL_UnlockAudio()

    def get_settings(self):
        """
        Gets the current audio interface settings. This method is only intended for testing purposes.
        Returns:
            A dictionary containing the current audio interface settings or None if the
            audio interface is not enabled.
        """
        if self.enabled:
            return {'sample_rate': self.audio_callback_data.sample_rate,
                    'audio_channels': self.audio_callback_data.channels,
                    'buffer_samples': self.audio_callback_data.buffer_samples,
                    'buffer_size': self.audio_callback_data.buffer_size,
                    'bytes_per_sample': self.audio_callback_data.bytes_per_sample,
                    'seconds_to_bytes_factor': self.audio_callback_data.seconds_to_bytes_factor
                    }
        else:
            return None

    cdef write_gst_log_message(self, message_type, message):
        """Write GStreamer log message to the mpfmc log"""
        # print(message_type, message)
        if message_type == 'error':
            self.log.error(message)
        elif message_type == 'warning':
            self.log.warning(message)
        elif message_type == 'info':
            self.log.info(message)

    @property
    def enabled(self):
        return Mix_GetMusicHookData() != NULL

    def enable(self):
        """
        Enables audio playback (begins audio processing)
        """
        self.log.debug("Enabling audio playback")
        Mix_HookMusic(self.audio_callback, &self.audio_callback_data)

    def disable(self):
        """
        Disables audio playback (stops audio processing)
        """
        self.log.debug("Disabling audio playback")
        self.stop_all_sounds()
        Mix_HookMusic(NULL, NULL)

    def shutdown(self):
        """
        Shuts down the audio device
        """
        self.disable()
        Mix_CloseAudio()

    @staticmethod
    def get_max_tracks():
        """ Returns the maximum number of tracks allowed. """
        return MAX_TRACKS

    @staticmethod
    def get_max_markers():
        """Return the maximum number of markers allowed per sound"""
        return MAX_MARKERS

    def get_track_count(self):
        """Returns the number of tracks that have been created."""
        return len(self.tracks)

    def get_track_names(self):
        """Return the list of names of tracks that have been created."""
        return [track.name for track in self.tracks]

    def get_track(self, int track_num):
        """
        Returns the track with the specified track number.
        Args:
            track_num: The track number to retrieve
        """
        try:
            return self.tracks[track_num]
        except IndexError:
            return None

    def get_track_type(self, str name not None):
        """
        Returns the type of the track with the specified name.
        Args:
            name: The name of the track
        """
        track = self.get_track_by_name(name)
        if track:
            return track.type
        else:
            return None

    def get_track_by_name(self, str name not None):
        """
        Returns the track with the specified name.
        Args:
            name: The track name to retrieve
        """
        name = name.lower()
        for track in self.tracks:
            if name == track.name:
                return track

        return None

    def create_standard_track(self, object mc, str name not None,
                              int max_simultaneous_sounds=MAX_SIMULTANEOUS_SOUNDS_DEFAULT,
                              float volume=1.0):
        """
        Creates a new standard track in the audio interface
        Args:
            mc: The media controller app
            name: The name of the new track
            max_simultaneous_sounds: The maximum number of sounds that may be played at one time on the track
            volume: The track volume (0.0 to 1.0)

        Returns:
            A Track object for the newly created track
        """
        cdef int track_num = len(self.tracks)
        if track_num == MAX_TRACKS:
            self.log.error("Add track failed - the maximum number of tracks "
                           "(%d) has been reached.", MAX_TRACKS)
            return None

        # Make sure track name does not already exist (no duplicates allowed)
        name = name.lower()
        for track in self.tracks:
            if name == track.name:
                self.log.error("Add track failed - the track name '%s' already exists.", name)
                return None

        # Make sure audio callback function cannot be called while we are changing the track data
        SDL_LockAudio()

        # Create the new standard track
        new_track = TrackStandard(mc,
                                  pycapsule.PyCapsule_New(&self.audio_callback_data, NULL, NULL),
                                  name,
                                  track_num,
                                  self.audio_callback_data.buffer_size,
                                  max_simultaneous_sounds,
                                  volume)
        self.tracks.append(new_track)

        # Update audio callback data with new track
        self.audio_callback_data.track_count = len(self.tracks)
        self.audio_callback_data.tracks[track_num] = new_track.state

        # Allow audio callback function to be called again
        SDL_UnlockAudio()

        self.log.debug("The '%s' standard track has successfully been created.", name)

        return new_track

    def create_sound_loop_track(self, object mc, str name not None,
                                int max_layers=8,
                                float volume=1.0):
        """
        Creates a new sound loop track in the audio interface
        Args:
            mc: The media controller app
            name: The name of the new track
            max_layers: The maximum number of sounds that may be played at one time on the track
            volume: The track volume (0.0 to 1.0)

        Returns:
            A Track object for the newly created track
        """
        cdef int track_num = len(self.tracks)
        if track_num == MAX_TRACKS:
            self.log.error("Add track failed - the maximum number of tracks "
                           "(%d) has been reached.", MAX_TRACKS)
            return None

        # Make sure track name does not already exist (no duplicates allowed)
        name = name.lower()
        for track in self.tracks:
            if name == track.name:
                self.log.error("Add track failed - the track name '%s' already exists.", name)
                return None

        # Make sure audio callback function cannot be called while we are changing the track data
        SDL_LockAudio()

        # Create the new live loop track
        new_track = TrackSoundLoop(mc,
                                   pycapsule.PyCapsule_New(&self.audio_callback_data, NULL, NULL),
                                   name,
                                   track_num,
                                   self.audio_callback_data.buffer_size,
                                   max_layers,
                                   volume)
        self.tracks.append(new_track)

        # Update audio callback data with new track
        self.audio_callback_data.track_count = len(self.tracks)
        self.audio_callback_data.tracks[track_num] = new_track.state

        # Allow audio callback function to be called again
        SDL_UnlockAudio()

        self.log.debug("The '%s' live loop track has successfully been created.", name)

        return new_track

    def create_playlist_track(self, object mc, str name not None, float crossfade_time=0.0, float volume=1.0):
        """
        Creates a new playlist track in the audio interface
        Args:
            mc: The media controller app
            name: The name of the new track
            crossfade_time: The default crossfade time (secs) to use when fading between sounds
            volume: The track volume (0.0 to 1.0)

        Returns:
            A Track object for the newly created track

        Notes:
            There is no specialized playlist track type.  This function creates a standard
            audio track AND a corresponding playlist controller that will be used to manage
            all sound actions on the track.
        """
        cdef int track_num = len(self.tracks)
        if track_num == MAX_TRACKS:
            self.log.error("Add track failed - the maximum number of tracks "
                           "(%d) has been reached.", MAX_TRACKS)
            return None

        # Make sure track name does not already exist (no duplicates allowed)
        name = name.lower()
        for track in self.tracks:
            if name == track.name:
                self.log.error("Add track failed - the track name '%s' already exists.", name)
                return None

        # Make sure audio callback function cannot be called while we are changing the track data
        SDL_LockAudio()

        # Create the new standard track
        new_track = TrackStandard(mc,
                                  pycapsule.PyCapsule_New(&self.audio_callback_data, NULL, NULL),
                                  name,
                                  track_num,
                                  self.audio_callback_data.buffer_size,
                                  2,
                                  volume)
        self.tracks.append(new_track)

        # Update audio callback data with new track
        self.audio_callback_data.track_count = len(self.tracks)
        self.audio_callback_data.tracks[track_num] = new_track.state

        # Allow audio callback function to be called again
        SDL_UnlockAudio()

        # Create playlist controller for the track
        playlist_controller = PlaylistController(mc, new_track, crossfade_time)
        self.playlist_controllers[name] = playlist_controller

        self.log.debug("The '%s' playlist track and controller have successfully been created.", name)

        return new_track

    def get_playlist_controller_count(self):
        """Returns the number of playlist controllers that have been created."""
        return len(self.playlist_controllers)

    def get_playlist_controller_names(self):
        """Return the list of names of the playlist controllers that have been created."""
        return [controller for controller in self.playlist_controllers.keys()]

    def get_playlist_controller(self, str controller_name):
        """
        Returns the playlist controller with the specified name.
        Args:
            controller_name: The playlist controller name to retrieve
        """
        try:
            return self.playlist_controllers[controller_name]
        except KeyError:
            return None

    def load_sound_file_to_memory(self, str file_name):
        """
        Loads an audio file into a SoundMemoryFile wrapper object for use in a Sound object.
        Used in asset loading for Sound objects.
        Args:
            file_name: The audio file name to load.

        Returns:
            A SoundMemoryFile wrapper object containing a pointer to the sound sample
            data in memory.  An exception is thrown if the sound is unable to be loaded.
        """
        return SoundMemoryFile(file_name, pycapsule.PyCapsule_New(&self.audio_callback_data, NULL, NULL))

    def load_sound_file_for_streaming(self, str file_name):
        """
        Loads an audio file into a SoundMemoryFile wrapper object for use in a Sound object.
        Used in asset loading for Sound objects.
        Args:
            file_name: The audio file name to load.

        Returns:
            A SoundMemoryFile wrapper object containing a pointer to the sound sample
            data in memory.  An exception is thrown if the sound is unable to be loaded.
        """
        return SoundStreamingFile(file_name, pycapsule.PyCapsule_New(&self.audio_callback_data, NULL, NULL))

    def unload_sound_file(self, container not None):
        """
        Unloads the source sample from the supplied container (used in Sound
        asset unloading).  The sound will no longer be in memory.
        Args:
            container: A SoundFile object
        """
        if not isinstance(container, SoundFile):
            return

        container.unload()

    def stop_all_sounds(self, float fade_out_seconds = 0.0):
        """Stops all playing and pending sounds in all tracks"""
        for track in self.tracks:
            track.stop_all_sounds(fade_out_seconds)

    def stop_sound_instance(self, sound_instance not None, fade_out=None):
        """Stops the specified sound instance"""
        for track in self.tracks:
            if hasattr(track, "stop_sound_instance"):
                track.stop_sound_instance(sound_instance, fade_out)

    def stop_sound(self, sound not None, fade_out=None):
        """Stops all instances of the specified sound on all tracks"""
        for track in self.tracks:
            if hasattr(track, "stop_sound"):
                track.stop_sound(sound, fade_out)

    def stop_sound(self, sound not None):
        """Stops all instances of the specified sound from continuing to loop on all tracks"""
        for track in self.tracks:
            if hasattr(track, "stop_sound_looping"):
                track.stop_sound_looping(sound)

    def stop_sound_instance_looping(self, sound_instance not None):
        """Stops the specified sound instance from continuing to loop."""
        for track in self.tracks:
            if hasattr(track, "stop_sound_instance_looping"):
                track.stop_sound_instance_looping(sound_instance)

    def clear_context(self, context):
        """Clears the context in all tracks"""
        for track in self.tracks:
            track.clear_context(context)

    def process(self):
        """Process tick function for the audio interface."""
        for track in self.tracks:
            track.process()

    @staticmethod
    cdef void audio_callback(void* data, Uint8 *output_buffer, int length) nogil:
        """
        Main audio callback function (called from SDL_mixer).
        Args:
            data: A pointer to the AudioCallbackData class for the channel (contains all audio
                processing-related settings and state, ex: interface settings, tracks, sound
                players, ducking envelopes, etc.)
            output_buffer: A pointer to the audio data buffer for SDL2 to process
            length: The length of the audio buffer (in bytes)

        Notes:
            This static function is responsible for filling the supplied audio buffer with sound.
            samples. The function is called during the custom music player callback. Individual
            track buffers are maintained in each Track object and are processed during this callback.
        """
        cdef Uint32 buffer_length = <Uint32> length
        cdef AudioCallbackData *callback_data = <AudioCallbackData*> data
        cdef TrackState *track

        if callback_data == NULL:
            return

        # Initialize master output buffer with silence as it arrives uninitialized
        memset(output_buffer, 0, buffer_length)

        # Note: There are three separate loops over the tracks that must remain separate due
        # to various track parameters than can be set for any track during each loop.  Difficult
        # to debug logic errors can occur when these track loops are combined.

        # Loop over tracks, initializing the status, track buffer, and track ducking.
        for track_num in range(callback_data.track_count):
            track = <TrackState*>callback_data.tracks[track_num]

            track.active = False
            memset(track.buffer, 0, buffer_length)

            track.ducking_is_active = False
            for control_point in range(CONTROL_POINTS_PER_BUFFER):
                g_array_set_val_uint8(track.ducking_control_points, control_point, SDL_MIX_MAXVOLUME)

        # Loop over tracks, mixing the playing sounds into the track's audio buffer
        for track_num in range(callback_data.track_count):
            track = <TrackState*>callback_data.tracks[track_num]

            # No need to process/mix the track if the track is stopped or paused
            if track.status == track_status_stopped or track.status == track_status_paused:
                continue

            # Call the track's mix callback function (generates audio into track buffer)
            if track.mix_callback_function != NULL:
                track.mix_callback_function(track, buffer_length, callback_data)

        # Loop over tracks again, applying ducking and mixing down tracks to the master output buffer
        for track_num in range(callback_data.track_count):
            track = <TrackState*>callback_data.tracks[track_num]

            # Only mix the track to the master output and apply ducking if it is active
            if track.active:
                Track.mix_track_to_output(track,
                                          callback_data,
                                          output_buffer,
                                          buffer_length)

        # Apply master volume to output buffer
        Track.apply_volume(output_buffer, output_buffer, buffer_length, callback_data.master_volume)



"""Starts the MPF media controller."""

import argparse
import logging
import os
import socket
import sys
import threading
from datetime import datetime
import time
import errno
import psutil
from mpf.core.config_loader import YamlMultifileConfigLoader, ProductionConfigLoader

from mpf.commands.logging_formatters import JSONFormatter


# Note, other imports are done deeper in this file, which we need to do there
# since Kivy does so much with singletons and we don't want MPF to import
# them when it reads this command


class Command:

    # pylint: disable-msg=too-many-locals
    # pylint: disable-msg=too-many-statements
    def __init__(self, mpf_path, machine_path, args):   # noqa
        """Run MC."""
        p = psutil.Process(os.getpid())
        # increase priority slightly. this will keep MPF responsive when MC lags
        if sys.platform == "win32":
            p.nice(psutil.BELOW_NORMAL_PRIORITY_CLASS)
        else:
            p.nice(10)
        # undo all of Kivy's built-in logging so we can do it our way
        os.environ['KIVY_NO_FILELOG'] = '1'
        os.environ['KIVY_NO_CONSOLELOG'] = '1'
        from kivy.logger import Logger

        for handler in Logger.handlers:
            Logger.removeHandler(handler)
        sys.stderr = sys.__stderr__

        # Need to have these in here because we don't want them to load when
        # the module is loaded as an mpf.command
        from mpf.core.utility_functions import Util

        del mpf_path

        parser = argparse.ArgumentParser(description='Starts the MPF Media Controller')

        parser.add_argument("-b",
                            action="store_false", dest="bcp", default=True,
                            help="Do not set up the BCP server threads")

        parser.add_argument("-c",
                            action="store", dest="configfile",
                            default="config", metavar='config_file(s)',
                            help="The name of a config file to load. Default is "
                                 "config.yaml. Multiple files can be used via a comma-"
                                 "separated list (no spaces between)")

        parser.add_argument("-C",
                            action="store", dest="mcconfigfile",
                            default="mcconfig.yaml",
                            metavar='config_file',
                            help="The MPF framework default config file. Default is "
                                 "<mpf-mc install folder>/mcconfig.yaml")

        parser.add_argument("-f",
                            action="store_true", dest="force_assets_load", default=False,
                            help="Load all assets upon startup. Useful for "
                            "ensuring all assets are set up properly "
                            "during development.")

        parser.add_argument("-l",
                            action="store", dest="logfile",
                            metavar='file_name',
                            default=os.path.join("logs", datetime.now().strftime(
                                "%Y-%m-%d-%H-%M-%S-mc-" + socket.gethostname() +
                                ".log")),
                            help="The name (and path) of the log file")

        parser.add_argument("-L",
                            action="store", dest="mc_logfile",
                            metavar='file_name',
                            default=None,
                            help="The name (and path) of the log file")

        parser.add_argument("--json-logging",
                            action="store_true", dest="jsonlogging",
                            default=False,
                            help="Enables json logging to file. ")

        parser.add_argument("-p",
                            action="store_true", dest="pause", default=False,
                            help="Pause the terminal window on exit. Useful "
                            "when launching in a separate window so you can "
                            "see any errors before the window closes.")

        parser.add_argument("-P",
                            action="store_true", dest="production", default=False,
                            help="Production mode. Will suppress errors, wait for hardware on start and "
                                 "try to exit when startup fails. Run this inside a loop.")

        parser.add_argument("-t",
                            action="store_false", dest='text_ui', default=True,
                            help="Use the ASCII text-based UI")

        parser.add_argument("--both",
                            action="store_true", dest='mpf_both', default=False)

        parser.add_argument("-v",
                            action="store_const", dest="loglevel", const=logging.DEBUG,
                            default=logging.INFO, help="Enables verbose logging to the"
                                                       " log file")

        parser.add_argument("-V",
                            action="store_true", dest="consoleloglevel",
                            default=logging.INFO,
                            help="Enables verbose logging to the console. Do NOT on "
                                 "Windows platforms")

        parser.add_argument("-a",
                            action="store_true", dest="no_load_cache",
                            help="Forces the config to be loaded from files "
                                 "and not cache")

        parser.add_argument("-A",
                            action="store_false", dest="create_config_cache",
                            help="Does not create the cache config files")

        # The following are just included for full compatibility with mpf.py
        # which is needed when using "mpf both".

        parser.add_argument("-x",
                            action="store_const", dest="force_platform",
                            const='virtual', help=argparse.SUPPRESS)

        parser.add_argument("--vpx",
                            action="store_const", dest="force_platform",
                            const='virtual_pinball', help=argparse.SUPPRESS)

        parser.add_argument("-X",
                            action="store_const", dest="force_platform",
                            const='smart_virtual', help=argparse.SUPPRESS)

        parser.add_argument("--no-sound",
                            action="store_true", dest="no_sound", default=False)

        args = parser.parse_args(args)

        args.configfile = Util.string_to_event_list(args.configfile)

        # Configure logging. Creates a logfile and logs to the console.
        # Formatting options are documented here:
        # https://docs.python.org/2.7/library/logging.html#logrecord-attributes

        try:
            os.makedirs(os.path.join(machine_path, 'logs'))
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                raise

        if args.mc_logfile:
            args.logfile = args.mc_logfile

        full_logfile_path = os.path.join(machine_path, args.logfile)

        try:
            os.remove(full_logfile_path)
        except OSError:
            pass

        # initialise file log
        if args.jsonlogging:
            formatter = JSONFormatter()
            file_log = logging.FileHandler(full_logfile_path)
            file_log.setFormatter(formatter)
            logging.getLogger('').addHandler(file_log)
        else:
            logging.basicConfig(level=args.loglevel,
                                format='%(asctime)s : %(name)s : %(message)s',
                                filename=full_logfile_path,
                                filemode='a')

        # define a Handler which writes INFO messages or higher to the
        # sys.stderr

        if args.text_ui and args.mpf_both:
            console = logging.NullHandler()
            console.setLevel(logging.ERROR)
        else:
            console = logging.StreamHandler()
            console.setLevel(args.consoleloglevel)

        # set a format which is simpler for console use
        formatter = logging.Formatter('%(name)s: %(message)s')

        # tell the handler to use this format
        console.setFormatter(formatter)

        # add the handler to the root logger
        logging.getLogger('').addHandler(console)

        from mpfmc.core.mc import MpfMc

        logging.info("Loading MPF-MC controller")

        thread_stopper = threading.Event()

        # load config
        if not args.production:
            config_loader = YamlMultifileConfigLoader(machine_path, args.configfile,
                                                      not args.no_load_cache, args.create_config_cache)
        else:
            config_loader = ProductionConfigLoader(machine_path)

        config = config_loader.load_mc_config()

        try:
            MpfMc(options=vars(args),
                  config=config,
                  thread_stopper=thread_stopper).run()
            logging.info("MC run loop ended.")
        except Exception as e:  # noqa
            logging.exception(str(e))

        logging.info("Stopping child threads... (%s remaining)", len(threading.enumerate()) - 1)

        thread_stopper.set()

        while len(threading.enumerate()) > 1:
            time.sleep(.1)

        logging.info("All child threads stopped.")
        logging.shutdown()

        if args.pause:
            input('Press ENTER to continue...')

        sys.exit()


def get_command():
    return 'mc', Command



import logging
from mpfmc.tests.MpfMcTestCase import MpfMcTestCase


class TestAudioDisabled(MpfMcTestCase):
    """
    Tests the default audio settings (no 'sound_system' entries in the config file)
    """

    def get_machine_path(self):
        return 'tests/machine_files/audio'

    def get_config_file(self):
        return 'test_audio_disabled.yaml'

    def test_default_sound_system(self):
        """ Tests the sound system and audio interface with when the config settings
        disable the sound system. """

        self.assertFalse(self.mc.machine_config['sound_system']['enabled'])
        self.assertIsNone(self.mc.sound_system)



"""An ellipse widget."""
from typing import Optional
from kivy.graphics.vertex_instructions import Ellipse as KivyEllipse
from kivy.graphics.context_instructions import Color, Rotate, Scale
from kivy.properties import NumericProperty

from mpfmc.uix.widget import Widget

MYPY = False
if MYPY:   # pragma: no cover
    from mpfmc.core.mc import MpfMc


class Ellipse(Widget):

    """An ellipse widget."""

    widget_type_name = 'Ellipse'
    animation_properties = ('x', 'y', 'width', 'pos', 'height', 'size', 'color',
                            'angle_start', 'angle_end', 'opacity', 'rotation', 'scale')
    merge_settings = ('width', 'height')

    def __init__(self, mc: "MpfMc", config: dict, key: Optional[str] = None, **kwargs) -> None:
        del kwargs
        super().__init__(mc=mc, config=config, key=key)

        # Bind to all properties that when changed need to force
        # the widget to be redrawn
        self.bind(pos=self._draw_widget,
                  size=self._draw_widget,
                  color=self._draw_widget,
                  rotation=self._draw_widget,
                  scale=self._draw_widget,
                  segments=self._draw_widget,
                  angle_start=self._draw_widget,
                  angle_end=self._draw_widget)

        self._draw_widget()

    def _draw_widget(self, *args) -> None:
        del args

        if self.canvas is None:
            return

        anchor = (self.x - self.anchor_offset_pos[0], self.y - self.anchor_offset_pos[1])
        self.canvas.clear()

        with self.canvas:
            Color(*self.color)
            Rotate(angle=self.rotation, origin=anchor)
            Scale(self.scale).origin = anchor
            KivyEllipse(pos=self.pos, size=self.size,
                        segments=self.segments,
                        angle_start=self.angle_start,
                        angle_end=self.angle_end)

    #
    # Properties
    #

    segments = NumericProperty(180)
    '''Defines how many segments will be used for drawing the ellipse. The
    drawing will be smoother if you have many segments.
    '''

    angle_start = NumericProperty(0)
    '''Specifies the starting angle, in degrees, of the disk portion of
    the ellipse.
    '''

    angle_end = NumericProperty(360)
    '''Specifies the ending angle, in degrees, of the disk portion of
    the ellipse.
    '''

    rotation = NumericProperty(0)
    scale = NumericProperty(1.0)


widget_classes = [Ellipse]



"""Widget showing an image."""
from typing import Optional, Union

from kivy.properties import ObjectProperty, NumericProperty, AliasProperty
from kivy.graphics import Rectangle, Color, Rotate, Scale

from mpfmc.uix.widget import Widget

MYPY = False
if MYPY:   # pragma: no cover
    from mpfmc.core.mc import MpfMc
    from mpfmc.assets.image import ImageAsset


class ImageWidget(Widget):

    """Widget showing an image."""

    widget_type_name = 'Image'
    merge_settings = ('height', 'width')
    animation_properties = ('x', 'y', 'color', 'rotation', 'scale', 'fps', 'current_frame', 'opacity')

    def __init__(self, mc: "MpfMc", config: dict, key: Optional[str] = None, **kwargs) -> None:
        super().__init__(mc=mc, config=config, key=key)
        self.size = (0, 0)

        self._image = None  # type: ImageAsset
        self._current_loop = 0

        # Retrieve the specified image asset to display.  This widget simply
        # draws a rectangle using the texture from the loaded image asset to
        # display the image. Scaling and rotation is handled by the Scatter
        # widget.
        image = None
        try:
            image = self.mc.images[self.config['image']]
        except KeyError:

            try:
                image = self.mc.images[kwargs['play_kwargs']['image']]
            except KeyError:
                pass

        if not image:
            if not self.mc.asset_manager.initial_assets_loaded:
                raise ValueError("Tried to use an image '{}' when the initial asset loading run has not yet been "
                                 "completed. Try to use 'init_done' as event to show your slides if you want to "
                                 "use images.".format(self.config['image']))

            raise ValueError("Cannot add Image widget. Image '{}' is not a "
                             "valid image name.".format(self.config['image']))

        # Updates the config for this widget to pull in any defaults that were
        # in the asset config
        self.merge_asset_config(image)

        if image.is_pool:
            self._image = image.get_next()
        else:
            self._image = image

        self._image.references += 1

        # If the associated image asset exists, that means it's loaded already.
        if self._image.image:
            self._image_loaded()
        else:
            # if the image asset isn't loaded, we set the size to 0,0 so it
            # doesn't show up on the display yet.
            # TODO Add log about loading on demand
            self.size = (0, 0)
            self._image.load(callback=self._image_loaded)

        # Bind to all properties that when changed need to force
        # the widget to be redrawn
        self.bind(pos=self._draw_widget,
                  color=self._draw_widget,
                  rotation=self._draw_widget,
                  scale=self._draw_widget)

    def __repr__(self) -> str:  # pragma: no cover
        try:
            return '<Image name={}, size={}, pos={}>'.format(self._image.name,
                                                             self.size,
                                                             self.pos)
        except AttributeError:
            return '<Image (loading...), size={}, pos={}>'.format(self.size,
                                                                  self.pos)

    def _image_loaded(self, *args) -> None:
        """Callback when image asset has been loaded and is ready to display."""
        del args

        # Setup callback on image 'on_texture' event (called whenever the image
        # texture changes; used mainly for animated images)
        self._image.image.bind(on_texture=self._on_texture_change)
        self._on_texture_change()

        self._draw_widget()

        # Setup animation properties (if applicable)
        if self._image.image.anim_available:
            self.fps = self.config['fps']
            self.loops = self.config['loops']
            if self.config['auto_play']:
                self.play()
            else:
                self.stop()

    def _on_texture_change(self, *args) -> None:
        """Update texture from image asset (callback when image texture changes)."""
        del args

        self.texture = self._image.image.texture
        self.size = self.texture.size
        self._draw_widget()

        # Handle animation looping (when applicable)
        ci = self._image.image
        if ci.anim_available and self.loops > -1 and ci.anim_index == len(ci.image.textures) - 1:
            self._current_loop += 1
            if self._current_loop > self.loops:
                ci.anim_reset(False)
                self._current_loop = 0

    def prepare_for_removal(self) -> None:
        """Prepare the widget to be removed."""
        super().prepare_for_removal()
        # stop any animations
        if self._image:
            self._image.references -= 1
            if self._image.references == 0:
                try:
                    self._image.image.anim_reset(False)
                # If the image was already unloaded from memory
                except AttributeError:
                    pass

    def _draw_widget(self, *args):
        """Draws the image (draws a rectangle using the image texture)"""
        del args

        anchor = (self.x - self.anchor_offset_pos[0], self.y - self.anchor_offset_pos[1])
        self.canvas.clear()

        with self.canvas:
            Color(*self.color)
            Rotate(angle=self.rotation, origin=anchor)
            Scale(self.scale).origin = anchor
            Rectangle(pos=self.pos, size=self.size, texture=self.texture)

    def play(self, start_frame: Optional[int] = 0):
        """Play the image animation (if images supports it)."""
        if start_frame:
            self.current_frame = start_frame

        # pylint: disable-msg=protected-access
        self._image.image._anim_index = start_frame
        self._image.image.anim_reset(True)

    def stop(self) -> None:
        """Stop the image animation."""
        self._image.image.anim_reset(False)

    #
    # Properties
    #

    def _get_image(self) -> Optional["ImageAsset"]:
        return self._image

    image = AliasProperty(_get_image)

    texture = ObjectProperty(None, allownone=True)
    '''Texture object of the image. The texture represents the original, loaded
    image texture.

    Depending of the texture creation, the value will be a
    :class:`~kivy.graphics.texture.Texture` or a
    :class:`~kivy.graphics.texture.TextureRegion` object.

    :attr:`texture` is an :class:`~kivy.properties.ObjectProperty` and defaults
    to None.
    '''

    loops = NumericProperty(-1)
    '''Number of loops to play then stop animating. -1 means keep animating.
    '''

    def _get_fps(self) -> Optional[float]:
        if self._image.image.anim_available:
            return int(1 / self._image.image.anim_delay)
        else:
            return None

    def _set_fps(self, value: float):
        if value > 0:
            self._image.image.anim_delay = 1 / float(value)
        else:
            self._image.image.anim_delay = -1

    fps = AliasProperty(_get_fps, _set_fps)
    '''The frames per second rate for the animation if the image is sequenced
    (like an animated gif). If fps is set to 0, the animation will be stopped.
    '''

    def _get_current_frame(self) -> int:
        return self._image.image.anim_index + 1

    def _set_current_frame(self, value: Union[int, float]):
        if not self._image.image.anim_available or not hasattr(self._image.image.image, 'textures'):
            return

        frame = (int(value) - 1) % len(self._image.image.image.textures)
        if frame == self._image.image.anim_index:
            return
        else:
            self._image.image._anim_index = frame   # pylint: disable-msg=protected-access
            self._image.image.anim_reset(True)

    current_frame = AliasProperty(_get_current_frame, _set_current_frame)
    '''The current frame of the animation.
    '''

    rotation = NumericProperty(0)
    '''Rotation angle value of the widget.

    :attr:`rotation` is an :class:`~kivy.properties.NumericProperty` and defaults to
    0.
    '''

    scale = NumericProperty(1.0)
    '''Scale value of the widget.

    :attr:`scale` is an :class:`~kivy.properties.NumericProperty` and defaults to
    1.0.
    '''


widget_classes = [ImageWidget]



from mpfmc.tests.MpfMcTestCase import MpfMcTestCase
from mpfmc.transitions.push import PushTransition
from kivy.uix.screenmanager import WipeTransition, NoTransition


class TestTransitions(MpfMcTestCase):
    def get_machine_path(self):
        return 'tests/machine_files/transitions'

    def get_config_file(self):
        return 'test_transitions.yaml'

    def test_transition_loading(self):
        # test that the mpf transitions have loaded
        self.assertIn('push', self.mc.transition_manager.transitions)
        self.assertEqual(self.mc.transition_manager.transitions['push'],
                         PushTransition)

        # test that the kivy transitions have loaded
        self.assertIn('wipe', self.mc.transition_manager.transitions)
        self.assertEqual(self.mc.transition_manager.transitions['wipe'],
                         WipeTransition)

    def test_mpf_transition(self):
        self.mc.events.post('show_slide1')

        # show a slide with a transition
        self.mc.events.post('show_slide2')
        self.advance_time(1)

        # transition is 2 secs, so we can check the progress here
        slide = self.mc.targets['default'].current_slide

        self.assertAlmostEqual(slide.transition_progress, 0.5, delta=.1)
        self.assertEqual(slide.transition_state,'in')

        transition = self.mc.targets['default'].transition
        self.assertEqual(transition.duration, 2.0)
        self.assertTrue(transition.is_active)
        self.assertEqual(transition.easing, 'out_bounce')
        self.assertTrue(isinstance(transition, PushTransition))

    def test_target_transition_reset_when_done(self):
        # Show slide1 with no transition
        self.mc.events.post('show_slide1')
        self.advance_time()
        self.assertTrue(isinstance(self.mc.targets['default'].transition,
                                   NoTransition))
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'slide1')

        # Show slide2 with a transition
        self.mc.events.post('push_left')
        self.advance_time()
        self.assertTrue(isinstance(self.mc.targets['default'].transition,
                                   PushTransition))
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'slide2')

        # Go back to slide1 with a transition
        self.mc.events.post('show_slide1_with_push')
        self.advance_time()
        self.assertTrue(isinstance(self.mc.targets['default'].transition,
                                   PushTransition))
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'slide1')

        self.mc.events.post('push_left')
        self.advance_time()
        self.assertTrue(isinstance(self.mc.targets['default'].transition,
                                   PushTransition))
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'slide2')

        self.mc.events.post('show_slide1')
        self.advance_time()
        self.assertTrue(isinstance(self.mc.targets['default'].transition,
                                   NoTransition))
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'slide1')

        self.mc.events.post('show_slide2_no_transition')
        self.advance_time()
        self.assertTrue(isinstance(self.mc.targets['default'].transition,
                                   NoTransition))
        self.assertEqual(self.mc.targets['default'].current_slide_name,
                         'slide2')

    def test_push_left(self):
        self.mc.events.post('show_slide1')
        self.advance_time()
        self.mc.events.post('push_left')
        self.advance_time()

    def test_push_right(self):
        self.mc.events.post('show_slide1')
        self.advance_time()
        self.mc.events.post('push_right')
        self.advance_time()

    def test_push_up(self):
        self.mc.events.post('show_slide1')
        self.advance_time()
        self.mc.events.post('push_up')
        self.advance_time()

    def test_push_down(self):
        self.mc.events.post('show_slide1')
        self.advance_time()
        self.mc.events.post('push_down')
        self.advance_time()

    def test_move_in_right(self):
        self.mc.events.post('show_slide1')
        self.advance_time()
        self.mc.events.post('move_in_right')
        self.advance_time()

    def test_move_in_left(self):
        self.mc.events.post('show_slide1')
        self.advance_time()
        self.mc.events.post('move_in_left')
        self.advance_time()

    def test_move_in_top(self):
        self.mc.events.post('show_slide1')
        self.advance_time()
        self.mc.events.post('move_in_top')
        self.advance_time()

    def test_move_in_bottom(self):
        self.mc.events.post('show_slide1')
        self.advance_time()
        self.mc.events.post('move_in_bottom')
        self.advance_time()

    def test_move_out_right(self):
        self.mc.events.post('show_slide1')
        self.advance_time()
        self.mc.events.post('move_out_right')
        self.advance_time()

    def test_move_out_left(self):
        self.mc.events.post('show_slide1')
        self.advance_time()
        self.mc.events.post('move_out_left')
        self.advance_time()

    def test_move_out_top(self):
        self.mc.events.post('show_slide1')
        self.advance_time()
        self.mc.events.post('move_out_top')
        self.advance_time()

    def test_move_out_bottom(self):
        self.mc.events.post('show_slide1')
        self.advance_time()
        self.mc.events.post('move_out_bottom')
        self.advance_time()

    def test_wipe(self):
        self.mc.events.post('show_slide1')
        self.advance_time()
        self.mc.events.post('wipe')
        self.advance_time()

    def test_swap(self):
        self.mc.events.post('show_slide1')
        self.advance_time()
        self.mc.events.post('swap')
        self.advance_time()

    def test_fade(self):
        self.mc.events.post('show_slide1')
        self.advance_time()
        self.mc.events.post('fade')
        self.advance_time()

    def test_fade_back(self):
        self.mc.events.post('show_slide1')
        self.advance_time()
        self.mc.events.post('fade_back')
        self.advance_time()

    def test_rise_in(self):
        self.mc.events.post('show_slide1')
        self.advance_time()
        self.mc.events.post('rise_in')
        self.advance_time()

    # def test_no_transition_1(self):
    #     self.mc.events.post('show_slide1')
    #     self.advance_time()
    #     self.mc.events.post('no_transition_1')
    #     self.advance_time(.5)
    #
    # def test_no_transition_2(self):
    #     self.mc.events.post('show_slide1')
    #     self.advance_time()
    #     self.mc.events.post('no_transition_2')
    #     self.advance_time(.5)
    #
    # def test_no_transition_3(self):
    #     self.mc.events.post('show_slide1')
    #     self.advance_time()
    #     self.mc.events.post('no_transition_3')
    #     self.advance_time(.5)
